/*
 * (Enhanced) Network block device - make block devices work over TCP
 *
 * Original NBD Copyright 1997 Pavel Machek <pavel@elf.mj.gts.cz>
 * Further ENBD Copyrights 1998, 1999, 2000 Peter Breuer <ptb@it.uc3m.es>
 *
 * Development of the ENBD software has been supported by grants and
 * contributions from Realm Information Technologies, Inc. of 5555
 * Oakbrook Parkway, NW Norcross, GA and iNsu Innovations Inc.  of
 * 3465, Boulevard Thimens, Saint-Laurent, Quebec, Canada.
 * 
 * ------------ Pavel's history notes ----------------------------------
 * 97-3-25 compiled 0-th version, not yet tested it 
 *   (it did not work, BTW) (later that day) HEY! it works!
 *   (bit later) hmm, not that much... 2:00am next day:
 *   yes, it works, but it gives something like 50kB/sec
 * 97-3-28 it's completely strange - when using 1024 byte "packets"
 *   it gives 50kB/sec and CPU idle; with 2048 bytes it gives
 *   500kB/sec (and CPU loaded 100% as it should be) (all done
 *   against localhost)
 * 97-4-1 complete rewrite to make it possible for many requests at 
 *   once to be processed
 * 97-4-1 23:57 rewrite once again to make it work :-(
 * 97-4-3 00:02 hmm, it does not work.
 * 97-4-3 23:06 hmm, it will need one more rewrite :-)
 * 97-4-10 It looks like it's working and stable. But I still do not
 *  have any recovery from lost connection...
 * (setq tab-width 4)
 * 97-4-11 Making protocol independent of endianity etc.
 * 97-4-15 Probably one more rewrite, since it loses requests under
 *  heavy loads
 * 97-9-13 Cosmetic changes
 *
 * possible FIXME: make set_sock / set_blksize / set_size / do_it one syscall
 * why not: would need verify_area and friends, would share yet another 
 *          structure with userland
 *
 * FIXME: not module-safe
 *
 * ------------ Peter's history notes ----------------------------------
 * 98-12-18 modules now OK ptb@it.uc3m.es (Peter Breuer) ported to
 * 2.0.*. + better debugging. Still possible lockup in connection with APM
 * and spurious interrupt - only on write. Error treatment should
 * be improved. After 100 errors from end_request the kernel can
 * do anything. We should catch it ourselves.
 * 99-1-sometime fixed lockup by extending semaphore - ptb v1.0.
 * 99-3-sometime reconnect protocol (client mod agreed by pavel) - ptb v1.1
 * 99-4-25 add /proc/nbdinfo - ptb v1.1.1
 * 99-4-sometime add multiplex - ptb v1.2
 * 99-4-26 fix multiplex and redundancy - ptb v1.2.1
 * 99-4-29 reentrant client threads - ptb v1.2.2
 * 99-4-29 socket related stuff placed in user space - amarin v1.3.0
 * 99-5-3  fix all, all writes had to be before all reads - ptb v1.2.4
 * 99-5-5  fix out-of-order, async - ptb v1.2.5
 * 99-5-7  semaphores removed (still works!), fail cases corrected - ptb v1.2.6
 * 99-5-12 signals unblocked in xmit, blksize != 1024 fixed, ioctls
 *         added  - ptb v1.2.7
 * 99-6-1  interaction with client split into two functions - amarin v1.3.0
 * 99-6-3  reintegrated fully, mem manager fixed, accounts fixed - ptb v1.2.8.3
 * 99-6-3  extra queue removed, mem manager removed  - ptb v1.2.8.4
 * 99-7-3  buffer registration introduced - ptb v1.2.8.5
 * 99-7-3  some client redundancy reestablished - ptb v2.1.1
 * 99-7-10 encapsulated queue calls. One element rollback buffer - ptb v2.1.2
 * 99-7-20 timestamp and rollback old abandoned request - ptb v2.1.3
 * 99-7-24 64bit file sizes and offsets accepted - ptb v2.1.9
 * 99-7-26 experimental request coalesces - ptb v2.1.10
 * 99-7-27 partitioning scheme - ptb v2.2.1
 * 99-8-3  nbd_clr_sock bug in invalidate_device fixed? - ptb v2.2.4
 * 99-8-5  reverse replace of block_fsync, add sig ioctls - ptb v2.2.5
 *         reverse bug introduced about v2.2.3 for compound reqs - ptb v2.2.5
 *         fix clear_que bug (didn't rollback first) from 2.1.3 - ptb v2.2.5
 * 99-8-22 workaround strange nr_sectors bug - ptb v2.2.6
 * 99-8-11 fix MY_NBD_SYNC bug. Never sync'ed all - ptb v2.2.7
 * 99-8-12 wakeups all moved to enqueue - ptb v2.2.7
 * 99-8-23 remove slot->cli_age - ptb v2.2.7
 * 99-8-24 first 8 bytes of signature embedded in packets - ptb v2.2.8
 *         fix SET_SIG define buglet, remove hardcoded constants - ptb v2.2.8
 *         fix huge bug. Missing copy_fromfs in my_nbd_ack - ptb v2.2.8     
 *         removed signature embedding and all other decorations - ptb v2.2.8
 * 99-8-25 recast fix in my_nbd_ack to avoid align. bug - ptb v2.2.9
 *         put in MKDEVs and put back some hardcode const fixes - ptb v2.2.10
 * 99-9-29 fix BLKGETSIZE bug - ptb v2.2.14
 * 99-10-2 run with interrupts on throughout. Think we lose some - ptb v2.2.15
 * 99-10-8 trim dead code, kernel 2.2 ifdef's - ptb v2.2.17
 * 99-12-18 further o-o - ptb v2.2.19
 * 99-12-28 queue account cleanup. endio on queue reqs at reset - ptb v2.2.20
 *          interruptible semaphores for better client recovery - ptb v2.2.20
 * 00-1-2   debugging cleanups. Fix race in end_request - ptb v2.2.21
 * 00-1-4   semaphores simplified. - ptb v2.2.22
 * 00-6-8   emergency control by write to proc - ptb v2.2.24
 * 00-7-20  ported to 2.4.0-test1. Possible minor bugs found/fixed - ptb v2.2.24
 * 00-7-27  changed proc i/f to read_proc from get_info in 2.2/2.4 - ptb v2.2.25
 * 00-7-30  fixed reads before writes under 2.4 by disabling merge - ptb v2.2.25
 * 00-7-30  and fixed merge_reqs for 2.4, now that I understand! - ptb v2.2.25
 * 00-7-30  fixed/introduced possible bug in end_io  for 2.2/2.4 - ptb v2.2.25
 * 00-7-30 fixed hitherto masked bug in read_stat in nbd_client - ptb v2.4.0
 * 00-7-30 added timeout to net writes in nbd_client - ptb v2.4.0
 * 00-8-20 display fix for devices over 2GB - ptb v2.4.5
 * 00-8-23 more 64 bit fixes + error out overrange requests- ptb v2.4.6/2.2.27
 * 00-8-31 add NBD_ERR ioctl to error out slot request- ptb v2.4.9
 * 00-8-31 soften NBD_SOFT_RESET so doesn't wreck protocol - ptb v2.4.9
 * 00-9-1  remove %L's from printfs. Kernel 2.2. doesn't - ptb v2.4.10/2.2.27
 * 00-9-6  add various state flags to help init order - ptb v2.4.11
 * 00-9-8  add checks for device initialised to set_sock - ptb v2.4.12
 * 00-9-17 en/disable device as aslot count goes through 0 - ptb v2.4.13/2.2.28
 */

static int paranoia = 1;
#define PARANOIA_BEGIN if(paranoia){
#define PARANOIA_END   }


#include <linux/major.h>

#include <linux/module.h>

#if defined(__GNUC__) && __GNUC__ >= 2
#define _LOOSE_KERNEL_NAMES
#endif

#include <linux/version.h>
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c)  (((a) << 16) + ((b) << 8) + (c))
#endif
#define LINUX_VERSION_2_2_0 KERNEL_VERSION(2,2,0)
#define LINUX_VERSION_2_1_0 KERNEL_VERSION(2,1,0)
#define LINUX_VERSION_2_3_30 KERNEL_VERSION(2,3,30)


#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/errno.h>

#include <asm/segment.h>
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
#include <asm/uaccess.h>          /* PTB - when did this arrive in kernel? */
#endif
#include <asm/byteorder.h>
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
#include <linux/wrapper.h>
#endif

#define MAJOR_NR NBD_MAJOR
#include <linux/proc_fs.h>
#include <linux/genhd.h>
#include <linux/hdreg.h>
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
#include <linux/file.h>           /* PTB - when did this arrive in kernel? */
#endif
#include <linux/nbd.h>

/* PTB kernel data - 4KB worth
 * We need space for nda, nda1, .. nda15, ndb, ndb1, ..
 * The index is exactly the minor number.
 */
static int nbd_blksizes [MAX_NBD * NBD_MAXCONN]; /* {1024, 1024,}; */
static int nbd_sizes    [MAX_NBD * NBD_MAXCONN]; /* {0x7fffffff, 0x7fffffff,};*/
static u64 nbd_bytesizes[MAX_NBD * NBD_MAXCONN]; /* all in init now PTB 132 */

/* PTB our data   - about 3KB
 * These are nda, ndb, ndc, ...
 * Divide the minor by NBD_MAXCONN to get this index.
 */
static struct nbd_device nbd_dev[MAX_NBD];

/* PTB compatibility */
#if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
  #define  DECLARE_MUTEX(x) struct semaphore x = MUTEX
  #define  init_MUTEX(x) { *(x) = MUTEX; }
  #define  init_MUTEX_LOCKED(x) { *(x) = MUTEX_LOCKED; }
  #define  blk_init_queue(device, fn) (device)->request_fn = (fn)
  #define  BLK_DEFAULT_QUEUE(major) (&blk_dev[major])
  #define  blkdev_dequeue_request(req) { \
     if (req==CURRENT) { \
          CURRENT=CURRENT->next; \
     } else { \
          struct request * r; \
          for (r = CURRENT; r && r->next; r = r->next) { \
            if (r->next == req)  { \
              r->next = req->next; \
              break; \
            } \
          } \
     } \
  }
  #define QUEUE_EMPTY (CURRENT==NULL)
  #define init_waitqueue_head(device)
  #define blk_cleanup_queue(device) (device)->request_fn = NULL
#endif
#if LINUX_VERSION_CODE < LINUX_VERSION_2_1_0
  #define MODULE_PARM(x,y)
  #define copy_from_user(x, y, z) memcpy_fromfs((x), (y), (z))
  #define copy_to_user(x, y, z) memcpy_tofs((x), (y), (z))
  #define fget(arg) (current->files->fd[(arg) & 0xffff])
  #define access_ok(vrw, buffer, size) (verify_area((vrw),(buffer),(size))==0)
  #define interruptible_sleep_on_timeout(wq, delay) { \
        current->timeout = (delay) + jiffies; \
        interruptible_sleep_on (wq); \
        current->timeout = 0; \
  }
  #define proc_register(x,y) proc_register_dynamic((x),(y))
#endif


/* PTB major request queue access */
DECLARE_MUTEX(nbd_queue_lock);

/* PTB list interface extensions */
#define list_head(ptr, type, member) \
  (list_empty(ptr)?NULL:list_entry(((struct list_head *)ptr)->next,type,member))
#define list_tail(ptr, type, member) \
  (list_empty(ptr)?NULL:list_entry(((struct list_head *)ptr)->prev,type,member))

#if LINUX_VERSION_CODE >= LINUX_VERSION_2_3_30
  #define NBD_LIST_ADD(req,lo) list_add(&req->queue, &lo->queue)
  #define NBD_LIST_DEL(req,lo) list_del(&req->queue)
  #define NBD_LIST_HEAD(lo) list_head(& lo->queue,struct request,queue)
  #define NBD_LIST_EMPTY(lo) list_empty(&lo->queue)
  #define NBD_LIST_ENTRY(pos) list_entry(pos,struct request,queue)
  #define NBD_LIST_FOR_EACH(pos,lo) list_for_each(pos, &lo->queue)
  #define NBD_LIST_NEXT(pos) list_head(& pos->queue,struct request,queue)
  #define NBD_LIST_TAIL(lo) list_tail(& lo->queue,struct request,queue)
  #define NBD_INIT_LIST_HEAD(lo) INIT_LIST_HEAD(& lo->queue)
#else
  /* PTB prepend. next points to front, prev points to back */
  #define NBD_LIST_ADD(req,lo) { \
      (struct request *) req->prev = lo->head; \
      (struct request *) req->next = NULL; \
      if (lo->head == NULL) { \
          lo->tail = lo->head = req; \
      } else { \
          lo->head->next = req; \
          lo->head = req; \
      } \
  }
  /* PTB remove. next points to front, prev points to back */
  #define NBD_LIST_DEL(req,lo) { \
      if (req->next) { \
        (struct request *) req->next->prev = req->prev; \
      } else { \
        lo->head = (struct request *) req->prev; \
      } \
      if (req->prev) { \
        ((struct request *)(req->prev))->next = req->next; \
      } else { \
        lo->tail = req->next; \
      } \
      req->next = NULL; \
      req->prev = NULL; \
  }
  #define NBD_LIST_HEAD(lo) (lo)->tail
  #define NBD_LIST_EMPTY(lo) (!(lo)->tail)
  #define NBD_LIST_ENTRY(pos) pos
  #define NBD_LIST_FOR_EACH(pos,lo) for (pos = lo->tail; pos; pos = pos->next)
  #define NBD_LIST_NEXT(pos) pos->next
  #define NBD_LIST_TAIL(lo) (lo)->tail
  #define NBD_INIT_LIST_HEAD(lo) lo->head = lo->tail = NULL
#endif

#if LINUX_VERSION_CODE >= LINUX_VERSION_2_3_30
  #define NBD_spin_unlock_irq(lock) spin_unlock_irq(lock)
  #define NBD_spin_lock_irq(lock) spin_lock_irq(lock)
#else
  #define NBD_spin_unlock_irq(lock)
  #define NBD_spin_lock_irq(lock)
#endif

#ifndef DEBUG
#define NDEBUG 0
#else
#define NDEBUG DEBUG
#endif

#define NBD_ID "NBD #%d[%d]: " __FUNCTION__ " "
#define NBD_DEBUG(level, s...) \
 if((NDEBUG)>=(level)&&debug)   { static int icnt; printk( KERN_DEBUG NBD_ID, __LINE__, icnt++); printk(s);}

#define NBD_ERROR( s...) { static int icnt; printk( KERN_ERR   NBD_ID, __LINE__, icnt++); printk(s);}
#define NBD_ALERT( s...) { static int icnt; printk( KERN_ALERT NBD_ID, __LINE__, icnt++); printk(s);}
#define NBD_INFO( s...)  { static int icnt; printk( KERN_INFO  NBD_ID, __LINE__, icnt++); printk(s);}

#define NBD_FAIL( s ) { \
  NBD_DEBUG(1, s " (result %d).\n" , result ); \
  goto error_out; \
}
#define NBD_HARDFAIL( s ) { \
  NBD_ERROR( s " (result %d).\n" , result ); \
  lo->harderror = result; \
  goto hard_error_out; \
}

/* PTB NBD semaphores with debug messages */
#define NBD_SEMAPHORE_UP(s) \
            { \
                up (s);  \
                NBD_DEBUG(2,"(queue) release semaphore %lx\n", (unsigned long)(s)); \
            }
#define NBD_SEMAPHORE_DOWN(s) \
            { \
                NBD_DEBUG(2,"(queue) will grab semaphore %lx\n" , (unsigned long)(s)); \
                down_interruptible (s); \
                NBD_DEBUG(2,"(queue) have grabbed semaphore %lx\n", (unsigned long)(s)); \
            }


/* PTB device parameters */
static int rahead     = NBD_RAHEAD;      /* PTB - read ahead blocks  */
static int sync_intvl = NBD_SYNC_INTVL;  /* PTB - client inactivity timeout */
static int merge_requests                /* PTB - bool, do request coalesce */
                      = NBD_MERGE_REQUESTS;
static int buf_sectors
                      = NBD_MAX_SECTORS; /* PTB - user bufsize required */
static int debug      = NDEBUG;

#if defined(MODULE) 
 MODULE_PARM(rahead,"i");
 MODULE_PARM(sync_intvl,"i");
 MODULE_PARM(merge_requests,"i");
 MODULE_PARM(buf_sectors,"i");
 MODULE_PARM(debug,"i");
#endif

#if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
#define prev sem                         /* PTB - a coarse hack */
#endif

/*
 * PTB count number of blocks in a request. This will be an overestimate
 * if the number is not an exact multiple. It seems to happen. We must
 * guarrantee to return 0 only if the request is invalid as we use this
 * test elsewhere.
 */
inline unsigned long
nr_blks(struct nbd_device *lo, struct request * req) {

    unsigned log_sectors_per_blk;
    unsigned sectors_per_blk;

    if (!req)
      return 0;

    log_sectors_per_blk = lo->logblksize - 9;
    sectors_per_blk = 1 << log_sectors_per_blk;

    return (req->nr_sectors  + sectors_per_blk - 1) >> log_sectors_per_blk;
}

/*
 * PTB - put a request onto the head of a nbd device's queue
 *     - presumably having taken it off the major's queue first!
 */
static void
nbd_enqueue (struct nbd_device *lo, struct request *req)
{
    unsigned long req_blks = nr_blks(lo,req);
    unsigned long flags;

    NBD_DEBUG (1, "(%d): entered\n", lo->islot);

    if (req_blks <= 0) {
        NBD_DEBUG (1, "(%d): exited no blocks in req\n", lo->islot);
        return;
    }


PARANOIA_BEGIN;
    if (1) {
      struct buffer_head *bh = req->bh;
      if (! test_bit(BH_Lock, &bh->b_state)) {
          NBD_ALERT("buffer head found unlocked!");
      }
    }
PARANOIA_END;

    #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
    req->next = NULL;
    #endif

    NBD_SEMAPHORE_DOWN(&lo->queue_lock);
    save_flags(flags); cli();         /* PTB irqs off - critical section */

    NBD_LIST_ADD(req, lo); /* PTB - really list_add(&req->queue, &lo->queue); */

    restore_flags(flags);             /* PTB end critical section */
    NBD_SEMAPHORE_UP(&lo->queue_lock);

    wake_up_interruptible(& lo->wq);

    /* PTB accounting */
    lo->requests_in += req_blks;
    lo->count++;
    if (lo->maxq < lo->count)
        lo->maxq = lo->count;

    NBD_DEBUG (1, "(%d): exited OK req with %lu blks added to device queue\n",
             lo->islot, req_blks);

}

#if 0
/*
 * PTB - pop a request off the tail of the nbd device general queue
 */
static struct request *
nbd_dequeue (struct nbd_device *lo)
{
    struct request *req = lo->tail;
    unsigned long flags;

    if (!req)
        return NULL;

    lo->count--;                /* PTB accounting */

    save_flags(flags); cli();         /* PTB irqs off - critical section */

    NBD_LIST_DEL(req,lo);

    restore_flags(flags);             /* PTB end critical section */

    return req;
}
#endif

/*
 * PTB - remove a request from anywhere in the nbd device general queue 
 *     - return 0 for success, -ve for fail
 */
static int
nbd_remove (struct nbd_device *lo, struct request *req)
{
    unsigned long flags;

    if (!req)
        return -2;

    save_flags (flags); cli();        

    /* PTB How can I check that the handle is valid first? */
    NBD_LIST_DEL(req,lo); /* PTB really list_del(&req->queue); */
    goto success;

  success:
    restore_flags (flags);
    lo->count--;                /* PTB accounting */
    return 0;
}

/*
 *  PTB - Open the device
 */
int
nbd_open (struct inode *inode, struct file *file)
{
    int dev = -1;
    struct nbd_device *lo = 0;
    int nbd = -1;

    NBD_DEBUG (3, "entered\n");

    if (1 && !inode && file) {  /* added by ptb for 2.0.35. Necessary? */
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
        inode = file->f_dentry->d_inode;
#else
        inode = file->f_inode;
#endif
    }
    if (!inode) {
        NBD_ERROR ("null inode.\n");
        NBD_DEBUG (3, "exited INVAL\n");
        return -EINVAL;
    }

    dev = MINOR (inode->i_rdev);
    nbd = dev >> NBD_SHIFT;

    NBD_DEBUG (3, "have inode %x for dev %c%d\n",
       (unsigned) inode, 'a' + nbd, dev);

    if (nbd >= MAX_NBD ) {
        NBD_ERROR ("too many (%d) whole devices open\n", nbd);
        NBD_DEBUG (3, "exited NODEV\n");
        return -ENODEV;
    }

    NBD_DEBUG (3, "have device %c%d\n", 'a' + nbd, dev);

    lo = &nbd_dev[nbd];

    NBD_DEBUG (3, "whole device %c has reference count %d\n",
       'a' + nbd, lo->refcnt);
 
    if (dev % NBD_MAXCONN == 0) {
      /* PTB we have got the whole dev's file or inode for 1st time */
      if (!lo->file || lo->file != file)  {
        NBD_DEBUG (3, "set file %x for whole device %c\n",
            (unsigned)file, 'a' + nbd );
         lo->file = file;
      }
      if (!lo->inode || lo->inode != inode) {
         NBD_DEBUG (3, "set inode %x for whole device %c\n",
            (unsigned)inode, 'a' + nbd );
         lo->inode = inode;
      }
    }

#if 0
    /* PTB possible sync before first opens */
    if (lo->refcnt <= lo->aslot|| ! MOD_IN_USE) {
       invalidate_buffers (inode->i_rdev);
       if (lo->inode)
         invalidate_buffers (lo->inode->i_rdev);
       else
         invalidate_buffers (MKDEV(NBD_MAJOR, nbd<<NBD_SHIFT));
       NBD_DEBUG (2, "invalidated buffers\n");
       fsync_dev (inode->i_rdev);
       if (lo->inode)
         fsync_dev (lo->inode->i_rdev);
       else
         fsync_dev (MKDEV(NBD_MAJOR, nbd<<NBD_SHIFT));
       NBD_DEBUG (2, "synced whole device %c\n", 'a' + nbd);
    }                                       
#endif

    lo->refcnt++;

    NBD_DEBUG (3, "increment device %c reference count to %d\n",
         'a' + nbd, lo->refcnt);
    MOD_INC_USE_COUNT;

    if (!(lo->flags & NBD_INITIALISED)) {       /* PTB 132 */
        NBD_DEBUG (1, "had to initialize unitialized device nd%c\n", 'a'+nbd);
        init_MUTEX(& lo->queue_lock);
        lo->flags |= NBD_INITIALISED;
        if (!(lo->flags & NBD_ENABLED)) {
          NBD_DEBUG (1, "had to enable disabled device nd%c\n", 'a'+nbd);
          lo->flags |= NBD_ENABLED;
        }
    }

    NBD_DEBUG (3, "exited OK\n");
    return 0;
}


/*
 * PTB - complete a transaction irrefutably by taking it out of the
 *     - slot pending position it is in, and reporting end_request to kernel
 */
void
nbd_commit (struct nbd_slot *slot, struct request * req)
{

    struct nbd_device *lo = slot->lo;
    unsigned long req_blks = nr_blks(lo,req);
    int errors;

    if (req_blks <= 0)
        return;

    errors = req->errors;

    NBD_LIST_DEL(req,slot); /* PTB really list_del(&req->queue); */

    nbd_end_request (req);

    lo->requests_req -= req_blks;

    slot->req_age = 0;
    if (errors <= 0) {
      slot->out += req_blks;
      lo->requests_out += req_blks;
    } else {
      slot->err += req_blks;
      lo->requests_err += req_blks;
    }
}




/*
 * PTB - undo transactions by taking them out of the slot pending
 *     - position and replacing them on the generic device queue
 */
void
nbd_rollback (struct nbd_slot *slot)
{

    struct nbd_device *lo = slot->lo;
    struct request *req;

    while (req = NBD_LIST_HEAD(slot), req) {

      unsigned long req_blks = nr_blks(lo,req);

      NBD_LIST_DEL(req,slot); /* PTB really list_del(&req->queue); */

      NBD_ALERT("rollback req %x from slot %d!\n" ,(unsigned) req, slot->i);

      /* PTB accounting */
      lo->requests_req -= req_blks;
      slot->in -= req_blks;
      lo->requests_in -= req_blks; /* PTB - pre-decrement count for enqueue */
      nbd_enqueue (lo, req);
    }

}

/*
 * PTB - error out a transaction irrefutably by taking it out of the
 *     - slot pending position it is in, and reporting end_request to kernel
 */
void
nbd_error (struct nbd_slot *slot)
{
    struct nbd_device *lo = slot->lo;
    struct request *req;

    while (req = NBD_LIST_HEAD(slot), req) {

      unsigned long req_blks = nr_blks(lo,req);

      req->errors++;

      NBD_LIST_DEL(req,slot); /* PTB really list_del(&req->queue); */
      NBD_ALERT("error out req %x from slot %d!\n" ,(unsigned) req, slot->i);

      nbd_end_request (req);

      lo->requests_req -= req_blks;
      slot->in -= req_blks;

      slot->req_age = 0;
      slot->err += req_blks;
      lo->requests_err += req_blks;
    }

}

/*
 * PTB - let a request onto the slot pending position
 */
void
nbd_accept(struct nbd_slot * slot, struct request * req) {

        struct nbd_device *lo = slot->lo;
        unsigned long req_blks = nr_blks(lo,req);

        if (req_blks <= 0)
            return;

        lo->requests_req += req_blks;
        NBD_LIST_ADD(req, slot); /* PTB - really list_add(&req->queue, &slot->queue); */
        slot->req_age = jiffies;
        slot->in += req_blks;
}

/*
 * PTB - andres' kernel half of the user-space network handshake, used
 *     - to complete a transaction.
 *     - return 0 for success and -ve for fail.
 */
int
nbd_ack (struct nbd_slot *slot)
{
    struct nbd_reply reply;
    struct request *req, *xreq;
    int result = 0;
    
    void *user;
    unsigned long req_blks = 1;
    struct nbd_device *lo = slot->lo;
    int islot = slot->i;

    NBD_DEBUG (1, "(%d): entered\n", islot);
    if (!(slot->flags & NBD_SLOT_BUFFERED)) {
      NBD_DEBUG (1, "(%d): exited INVAL\n", islot);
      return -EINVAL;
    }
    lo->cthreads++;
    slot->flags |= NBD_SLOT_RUNNING;

    user = slot->buffer + slot->buflen;

    copy_from_user ((char *) &reply, (char *) user, sizeof (struct nbd_reply));
    slot->buflen += sizeof (struct nbd_reply);

    memcpy ((char *) &req, reply.handle, sizeof (req));

    do {
    #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
      struct request * pos;
    #else
      struct list_head * pos;
    #endif
      int count = 0;
      xreq = NULL;
      NBD_LIST_FOR_EACH(pos, slot) {
        xreq = NBD_LIST_ENTRY(pos);
        NBD_DEBUG (3, "req %d on slot %d queue is %x\n",
                               count,slot->i,(unsigned)xreq);
        count++;
        if (xreq == req)
          /* PTB found it */
          break;
      }
    } while(0);

    if (xreq != req) {
        NBD_ALERT ("fatal: Bad handle %x !\n" ,(unsigned) req);
        result = -EAGAIN;
        NBD_FAIL ("exited wrong request\n");
        /* PTB we will roll back requests */
    }

    NBD_DEBUG (3, "(%d): ok, got handle %x\n", islot, (unsigned) req);

    if (reply.magic != NBD_REPLY_MAGIC) {
        NBD_ALERT ("Not enough magic in " __FUNCTION__ "\n");
        result = -EAGAIN;
        NBD_FAIL ("Not enough magic in " __FUNCTION__ "\n");
        /* PTB we will roll back requests */
    }

    if (reply.error > 0 || req->errors > 0) {
        /* PTB wasn't error++'ed before */
        req->errors++;
        NBD_ALERT ("exited with reply error\n");
        /* PTB we handle this internally */
        goto success;
    }

    PARANOIA_BEGIN
     if (lo->magic != NBD_DEV_MAGIC) {
        NBD_ALERT ("nbd_dev[] corrupted: Not enough magic\n");
        NBD_DEBUG (1, "(%d): exited bad magic\n", islot);
        result = -ENODEV;
        NBD_FAIL ("exited because of bad magic on device\n");
        /* PTB we will roll back requests */
    }
    PARANOIA_END

    req_blks = nr_blks(lo,req);

    switch (req->cmd & 0x03) {
      unsigned len;
      char * src;
 
      case READ:
        #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
          len = req->current_nr_sectors;
          len <<= 9;
        #else
          len = req->nr_sectors;
          len <<= 9;
        #endif
          if (req->nr_sectors > buf_sectors) {
            NBD_ALERT("req %x nr_sectors %ld bigger than buf %d)\n",
                  (unsigned)req,req->nr_sectors,buf_sectors);
            len = buf_sectors;
            len <<= 9;
          }
          src = ((char *) user) + sizeof(struct nbd_reply);
          NBD_DEBUG (3, "read %d bytes data from %x to %x\n",
                     len, (unsigned) src, (unsigned) req->buffer);

          copy_from_user (req->buffer, src, len);

          slot->buflen += len;

#if DEBUG > 0
          if (lo->blockmap) {

            unsigned long i, imax = len >> lo->logblksize;
            unsigned long blknr = req->sector >> (lo->logblksize - 9);

            for (i = 0; i < imax; i++) {
              int block = blknr + i;
              int offset = block >> 2;
              int shift = (block & 3) << 1;
              switch((lo->blockmap[offset] >> shift) & 3) {
                case READ: // PTB 0 = READ -> READ|2
                  lo->blockmap[offset] |= 2 << shift;
                  break;
                case READ|2:
                  break;
                case WRITE:
                default:  // PTB 3 = empty -> READ = 0
                  lo->blockmap[offset] &= ~(3 << shift);
                  break;
              }
            }
          }
#endif
          break;
      case WRITE:

#if DEBUG > 0
          if (lo->blockmap) {

            unsigned long i;
            unsigned long blknr = req->sector >> (lo->logblksize - 9);

            for (i = 0; i<req_blks; i++) {
              int block = blknr + i;
              int offset = block >> 2;
              int shift = (block & 3) << 1;
              lo->blockmap[offset] &= ~(3 << shift);
              lo->blockmap[offset] |= WRITE << shift;
            }
          }
#endif
          break;
    }

  success:
    /* PTB - completion (or erroring) of transaction */
    nbd_commit (slot, req);

    slot->buflen = 0;
    lo->cthreads--;
    slot->flags &= ~NBD_SLOT_RUNNING;
    NBD_DEBUG (1, "(%d): exited OK\n", islot);
    return 0;

  error_out:                    /* PTB we will do a client rollback */
    req->errors += req_blks;
    slot->err += req_blks;
    slot->buflen = 0;
    lo->cthreads--;
    slot->flags &= ~NBD_SLOT_RUNNING;
    result = result < 0 ? result : -ENODEV;
    NBD_DEBUG (1, "(%d): exited FAIL\n", islot);
    return result;
}


/*
 * PTB - apply repeatedly to tail of queue to try and merge successive requests
 *     - returns 1 for success and 0 for fail
 */
static inline int 
attempt_merge (struct nbd_device *lo, struct request *req)
{
    struct request *next = 0;

#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
    return 0;
#endif

    if (!req) {
        NBD_DEBUG (2, "no request\n");
        return 0;
    }
    next = NBD_LIST_NEXT(req);
    if (!next) {
        NBD_DEBUG (2, "no next request\n");
        return 0;
    }
    if (req->sector + req->nr_sectors != next->sector) {
        NBD_DEBUG (2, "sectors not adjacent\n");
        return 0;
    }
    if (req->cmd != next->cmd) {
        NBD_DEBUG (2, "request types differ\n");
        return 0;
    }
    if (req->rq_dev != next->rq_dev) {
        NBD_DEBUG (2, "request devices differ\n");
        return 0;
    }
    if (req->nr_sectors + next->nr_sectors >= buf_sectors) {
        NBD_DEBUG (2, "buffer would be too big\n");
        return 0;
    }
#if 1
    if (req->buffer + (req->nr_sectors << 9) != next->buffer) {
        NBD_DEBUG (2, "buffers %x %x not contiguous\n",
                 (unsigned) req->buffer, (unsigned) next->buffer);
        return 0;
    }
#endif

#if 0
    lo = &nbd_dev[MINOR (req->rq_dev) >> NBD_SHIFT];            /* PTB do this instead? */ 
#endif
#if 0
    NBD_DEBUG (1, "merge %ld, %ld + %ld == %ld\n",
               req->sector, req->nr_sectors, next->nr_sectors,
               req->nr_sectors + next->nr_sectors);
#endif
    if (merge_requests) {
        NBD_DEBUG (1, "will merge request\n");
        req->bhtail->b_reqnext = next->bh;
        req->bhtail = next->bhtail;
        req->nr_sectors += next->nr_sectors;
        next->rq_status = RQ_INACTIVE;          /* PTB FIXME!? */
        NBD_LIST_DEL(next,lo); /* PTB really list_del(&next->queue); */
        lo->count--;
        return 1;
    }
    else {
        NBD_DEBUG (1, "would merge request\n");
        return 0;
    }
}


/*
 * PTB - andres' kernel half of the userspace networking. This part
 *     - initiates the transaction by taking a request off the generic
 *     - device queue and placing it in the slots pending position.
 *     - I believe we return 0 for success and -ve for fail.
 */
int 
nbd_get_req (struct nbd_slot *slot)
{
    struct nbd_request request;
    struct request *req;
    /* unsigned long flags; */ /* PTB - in case we use cli/sti */
    int result = 0;
    static int count;
    unsigned start_time = jiffies;
    const unsigned timeout = sync_intvl * HZ;
    void *user;
    struct nbd_device *lo = slot->lo;
    int islot = slot->i;

    NBD_DEBUG (1, "(%d): entered\n", islot);
    if (!(slot->flags & NBD_SLOT_BUFFERED)) {
      NBD_FAIL ("Our slot has no buffer\n");
    }

    lo->cthreads++;             /* PTB - client thread enters */
    slot->flags |= NBD_SLOT_RUNNING;
    user = slot->buffer + slot->buflen;
    lo->islot = islot;

    if (! NBD_LIST_EMPTY(slot)) {
        NBD_FAIL ("already treating one request\n");
        /* PTB we do a nontrivial rollback from the user daemon */
    }
    if (! slot->file) {
        result = -EBADF;
        NBD_FAIL ("Our slot has been nofiled\n");
    }
    if (!(lo->flags & NBD_ENABLED)) {
        result = -ENODEV;
        NBD_FAIL ("Our slot has been vamooshed\n");
    }

    /* PTB - now spin until request arrives to treat */

    while (NBD_LIST_EMPTY(lo) && slot->file) {

        NBD_DEBUG (3, "(%d): going to sleep %d\n", islot, count);
        lo->cwaiters++;
        slot->flags |= NBD_SLOT_WAITING;

        interruptible_sleep_on_timeout(&lo->wq, start_time + timeout - jiffies);
        NBD_DEBUG (3, "(%d): I wake up %d\n", islot, count);

        slot->flags &= ~NBD_SLOT_WAITING;
        lo->cwaiters--;
        count++;

        /* PTB fail for recheck if we're inactive too long */

        if (jiffies >= start_time + timeout && NBD_LIST_EMPTY(lo)) {
            NBD_DEBUG (3, "(%d): I timeout in the %dth cycle %ld jiffies late\n"
                   ,islot, count, jiffies - (start_time + timeout));
            result = -ETIME;
            NBD_DEBUG (3, "ho hum nothing to do again, timeout -ETIME\n");
            if (1) {
                int siz = lo->blksize + sizeof(struct nbd_request);
                /* PTB verify the buffer is still OK - holds one block */
                if (!access_ok(VERIFY_WRITE, slot->buffer, siz)){
                    static int nbd_clr_sock (struct nbd_slot * slot);
                    result = -EINVAL;
                    nbd_clr_sock (slot); // PTB TEST
                    NBD_FAIL("Our process has died or lost its buffer\n");
                }
            }
            goto error_out;
            /* PTB we may do a trivial rollback from the user daemon */
        }

    }

    lo->cwaiters++;
    slot->flags |= NBD_SLOT_WAITING;
    NBD_SEMAPHORE_DOWN (&lo->queue_lock);
    lo->cwaiters--;
    slot->flags &= ~NBD_SLOT_WAITING;

    if (!(lo->flags & NBD_ENABLED)) {
        result = -ENODEV;
        NBD_FAIL ("Our slot vaporized while we slept!\n");
    }
    if (! slot->file ) {
        result = -EBADF;
        NBD_FAIL ("Our slot nofiled itself while we slept!\n");
    }

    if (NBD_LIST_EMPTY(lo)) {
        NBD_SEMAPHORE_UP (&lo->queue_lock);
        /* PTB - somebody else did it while we waited on semaphore. OK */
        result = -EINVAL;
        NBD_FAIL ("ho hum beaten to the punch\n");
        /* PTB we may do a trivial rollback from the user daemon */
    }

    if (! NBD_LIST_EMPTY(slot)) {
        NBD_SEMAPHORE_UP (&lo->queue_lock);
        result = -EINVAL;
        NBD_FAIL ("already treating one request\n");
        /* PTB we do a nontrivial rollback from the user daemon */
    }

    /* PTB cli/sti here looks unnec. hardware interrupts return here */
    /* AMARIN begin uninterruptable code */

    /* PTB oldest=last element in queue */
    req = NBD_LIST_TAIL(lo);

    PARANOIA_BEGIN;
    if (!req) {
        NBD_SEMAPHORE_UP (&lo->queue_lock);
        result = -EINVAL;
        NBD_FAIL ("tail request is zero although queue reported nonempty!\n");
    }
    PARANOIA_END;

#if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
    if (merge_requests) {
        /* PTB coalesce queue requests */
        unsigned long size = req->nr_sectors << 9;
        NBD_DEBUG (1, "(%d): req=0x%lx has size %lu\n",
                   islot, (unsigned long) req, size);


        while (attempt_merge (lo, req) > 0) ;
        size = req->nr_sectors << 9;
        NBD_DEBUG (1, "(%d): req=0x%lx has size %lu\n",
                   islot, (unsigned long) req, size);
    }
#endif
                              /* PTB - must succeed as have the semaphore */
    result = nbd_remove (lo, req);
                              /* PTB now holding irqs off in nbd_remove */

    NBD_DEBUG (2, "(%d):"
               " despatching request 0x%lx position 0x%lx in queue\n",
               islot, (unsigned long) req, (unsigned long) NBD_LIST_TAIL(lo));

    /* AMARIN end uninterruptable code */
    /* PTB uh - maybe cli/sti is needed? interrupts can muck the queue?
     *        - Nah! I've left them enabled so we can see any errors.
     */

    NBD_SEMAPHORE_UP (&lo->queue_lock);

    if (req->sector + req->nr_sectors > (lo->size << 1)) {
        result = -EINVAL;
        req->errors++;
        // PTB short circuit the protocol and error out now
        nbd_commit(slot,req);
        NBD_FAIL ("Overrange request\n");
    }

    request.magic = NBD_REQUEST_MAGIC;
    request.type = req->cmd;

    request.from = req->sector;
    request.from <<= 9;

    request.len = req->nr_sectors;
    request.len <<= 9;

    request.flags = 0;

    memcpy (request.handle, (char *) &req, sizeof (req));

    NBD_DEBUG (2, "(%d):"
               "magic=0x%x, type=%d, from=0x%x, len=%u, handle=0x%x, "
               "\n",
               islot,
               (unsigned) request.magic,
               (unsigned) request.type,
               (unsigned) request.from,
               (unsigned) request.len,
               (unsigned) *(char **) request.handle
               );

PARANOIA_BEGIN;
    /* PTB do not assume buffer OK, although verified by REG_BUF
     *     I half believe that we could write to a dead processes
     *     buffer?
     */

    if (!access_ok(VERIFY_WRITE, user, sizeof(struct nbd_request)) ){
        result = -EINVAL;
        NBD_DEBUG(3, "(%d): exited with %d\n" , islot, result );
        NBD_FAIL ("invalid user buffer\n");
        /* PTB we do a nontrivial rollback from the user daemon, if alive */
    }
    NBD_DEBUG (3, "(%d): verified user buf 0x%x\n", islot, (unsigned)user);
PARANOIA_END;

    copy_to_user (user, (char *) &request, sizeof (struct nbd_request));

    slot->buflen += sizeof (struct nbd_request);

    if ((request.type & 0x03) == WRITE) {

        unsigned size = 0;
        unsigned current_size = req->current_nr_sectors << 9;
        char  *buffer = req->buffer;
        struct buffer_head *bh = req->bh;

        /* PTB assume user verified */

        while (size < request.len) {
            struct buffer_head *next = bh->b_reqnext;
            unsigned long offset = sizeof (struct nbd_request) + size;

            NBD_DEBUG (3, "(%d): write %d to user at 0x%x\n"
                , islot, current_size, (unsigned)(((char*)user) + offset));

            copy_to_user (((char *)user) + offset, buffer, current_size);

            size += current_size;
            slot->buflen += current_size;

            /* PTB - arrgh. Subtle bug. If this request is partially
             * handled by the daemon which dies, we mark its buffers
             * up to date now and then roll it back later, to be handled
             * again. The second time nasty things happen. Don't.
             */

            //bh->b_reqnext = 0;          // PTB don't add !
            //mark_buffer_uptodate(bh,1); // PTB don't add !
            //unlock_buffer(bh);          // PTB don't add !

            bh = next;
            if (!bh)
                break;

            NBD_DEBUG (1, "(%d)"
            " bh has buffer 0x%lx len=%lu in req with buffer 0x%lx len=%lu\n",
                       islot,
                       (unsigned long) bh->b_data, (unsigned long)bh->b_size,
                       (unsigned long) req->buffer, req->nr_sectors << 9);
            buffer = bh->b_data;
            current_size = bh->b_size;

        }
        //req->bh = 0;                   // PTB don't add !
    }

    nbd_accept (slot, req);

    slot->buflen = 0;
    lo->cthreads--;             /* PTB - client thread leaves normally */
    slot->flags &= ~NBD_SLOT_RUNNING;
    NBD_DEBUG (1, "exiting\n");

    return 0;

  error_out:
    slot->buflen = 0;
    /* PTB accounting - a fail to get a request is not an errored request */
    lo->cthreads--;             /* PTB - client thread leaves abnormally */
    slot->flags &= ~NBD_SLOT_RUNNING;
    result = result < 0 ? result : -ENODEV;

    NBD_DEBUG (1, "(%d): error exit %d\n", islot, result);
    return result;
}

static int
nbd_clr_device () {

    int count = 0;
    NBD_DEBUG (1, "entered\n");
    while (1) {
        struct request *req;
        unsigned long flags;

        NBD_SEMAPHORE_DOWN (&nbd_queue_lock);
        save_flags(flags); cli();
        NBD_spin_lock_irq(&io_request_lock);
        req = CURRENT;
        if (!req) {
          NBD_spin_unlock_irq(&io_request_lock);
          restore_flags(flags);
          NBD_SEMAPHORE_UP (&nbd_queue_lock);
          break;
        }
        blkdev_dequeue_request(req);
        // PTB I don't know if b_d_r moves on the pointer or not.
        if (CURRENT==req) {
          //CURRENT = blkdev_next_request(req);
          NBD_ALERT ("CURRENT did not move after req removed!\n");
          break;
        }
        NBD_spin_unlock_irq(&io_request_lock);
        restore_flags(flags);
        NBD_SEMAPHORE_UP (&nbd_queue_lock);
        req->errors++;
        nbd_end_request(req);
        count++;
    }
    NBD_ALERT ("removed %d requests\n", count);
    NBD_DEBUG (1, "exited\n");
    return count;
}

/*
 * PTB error out the pending requests on the nbd full device
 */
void
nbd_clr_queue (struct nbd_device *lo)
{
    int count = 0;

    NBD_DEBUG (1, "entered\n");

    while (1) {

        struct request *req;
        int invalid = 0;
        unsigned long req_blks = 1;

        if (NBD_LIST_EMPTY(lo)) {
            NBD_DEBUG (1, "exited req=0\n");
            break;
        }
        req = NBD_LIST_HEAD(lo);

        req_blks = nr_blks(lo,req);

        PARANOIA_BEGIN;
        if (lo != &nbd_dev[MINOR (req->rq_dev)]) {
            NBD_ALERT ("request corrupted when clearing!\n");
            invalid = 1;
        }
        if (lo->magic != NBD_DEV_MAGIC) {
            NBD_ERROR ("Not enough magic when clearing!\n");
            break;
        }
        PARANOIA_END;

        req->errors += req_blks;
        lo->requests_err += req_blks;

        nbd_remove(lo,req);
                                  /* PTB now hold irqs off in nbd_remove */
        count++;
        if (!invalid) {
            nbd_end_request (req);
        }

    }
    count += nbd_clr_device();
    NBD_ALERT ("unqueued %d reqs\n", count);
    NBD_DEBUG (3, "exited\n");
}

/*
 * We always wait for result of write, for now. It would be nice to make it
 * optional in future
 * if ((req->cmd & 0x03 == WRITE) && (lo->flags & NBD_WRITE_NOCHK)) 
 *   { printk( "Warning: Ignoring result!\n"); nbd_end_request( req ); }
 */

#undef NBD_FAIL
#define NBD_FAIL( s ) { \
  NBD_ERROR( s "(device %x).\n" , dev ); \
  goto error_out; \
}

static void
#if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
nbd_do_request (void)
#else
nbd_do_request(request_queue_t * q)
#endif
{
    struct request *req;
    int dev;
    struct nbd_device *lo = 0;
    unsigned long flags;
    int nbd;

    NBD_DEBUG(1, "entered\n" );

    INIT_REQUEST;               /* PTB */

    while(1) {

        unsigned size = 0;
        unsigned long req_blks = 1;

        NBD_SEMAPHORE_DOWN (&nbd_queue_lock);
        save_flags(flags); cli();

        NBD_spin_lock_irq(&io_request_lock);
        if (QUEUE_EMPTY) {
            restore_flags(flags);
            NBD_spin_unlock_irq(&io_request_lock);
            NBD_SEMAPHORE_UP (&nbd_queue_lock); /**/
            goto success_out;
        }

        req = CURRENT;
        if (!req) {
            restore_flags(flags);
            NBD_spin_unlock_irq(&io_request_lock);
            NBD_SEMAPHORE_UP (&nbd_queue_lock); /**/
            goto success_out;
        }

        /* PTB - have request. Unqueue it and place on our device */

        /* PTB put me here (no block before change!) */
        blkdev_dequeue_request(req);
        NBD_spin_unlock_irq(&io_request_lock);
        restore_flags(flags);
        NBD_SEMAPHORE_UP (&nbd_queue_lock); /**/

        size = req->nr_sectors << 9;
        NBD_DEBUG (1, "req=%lx has size %u\n", (unsigned long)req, size);

        dev = MINOR (req->rq_dev);
        nbd = dev >> NBD_SHIFT;

        NBD_DEBUG(1, "request on minor %d\n" , dev) ; 
        lo = &nbd_dev[nbd];

        lo->kthreads++;         /* PTB - one kernel thread enters */
        if (lo->kthreads > lo->kmax)
           lo->kmax = lo->kthreads;
        NBD_DEBUG (3, "DEBUG 9\n");

        if (nbd >= MAX_NBD ){
           NBD_FAIL ("Minor too big in nbd_do_request.");
        }
        if ((req->cmd & 0x03) == WRITE && (lo->flags & NBD_READ_ONLY)){
           NBD_FAIL ("Write on read-only");
        }
        if (!(lo->flags & NBD_INITIALISED)) {
           NBD_FAIL ("Device not initialised.");
        }
        if (lo->nslot <= 0) {
           NBD_FAIL ("Device has no clients registered yet.");
        }
        //if (lo->aslot <= 0) {
        //   NBD_FAIL ("Device presently has no active clients.");
        //}
PARANOIA_BEGIN
        if (lo->magic != NBD_DEV_MAGIC){
           NBD_FAIL ("nbd[] is not magical!"); 
        }
PARANOIA_END;

        #if DEBUG == 5           /* PTB test nbd_do_request in isolation */
           NBD_FAIL ("testing discard request");  /* PTB FIXME */
        #endif

        req->errors = 0;
        nbd_enqueue(lo,req);

        lo->kthreads--;         /* PTB one kernel thread leaves normally */

        NBD_DEBUG (3, "DEBUG 5\n");
        continue;

        error_out:
        NBD_DEBUG (3, "DEBUG 6\n");
        req_blks = nr_blks(lo,req);
        req->errors  += req_blks;
        lo->requests_err += req_blks;
        NBD_ALERT ("ending req %x with prejudice\n", (unsigned)req);
        nbd_end_request(req);   /* PTB valid? */

        lo->kthreads--;         /* PTB one kernel thread leaves abnormally */
        NBD_DEBUG (3, "DEBUG 7\n");
        return;                 /* or continue? PTB */
    }

  success_out:
    NBD_DEBUG(1, "exiting\n") ; 
    return;
}

/*
 * PTB rollback all requests on a given slot and invalidate it.
 */
static int
   /* introduced by PTB for better modularity */
nbd_clr_sock (struct nbd_slot * slot)
   /* PTB arg introduced for multiplexing */
   /* PTB the nbd arg if -ve, means don't invalidate buffers also */
{
    int i = 0;
    struct nbd_device *lo = slot->lo;
    int islot = slot->i;

    NBD_DEBUG (3, "entered with arg %d\n", islot);

    //NBD_SEMAPHORE_DOWN (&lo->queue_lock);     /* PTB */

    nbd_rollback(slot);

    if (slot->file) {
        #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
        slot->file->f_count--;
        #else
        atomic_dec(& slot->file->f_count);
        #endif
        slot->file = NULL;
        slot->sock = NULL;
        slot->pid  = 0;
        slot->bufsiz = 0;
        slot->flags= 0;
        slot->buffer = NULL;
    }

    NBD_SEMAPHORE_DOWN (&lo->queue_lock);       /* PTB */

    /* PTB reset lo->aslot */

    if (lo->aslot > 0) {
        /* PTB grr .. do this the hard way since I don't seem to count right */
        lo->aslot = 0;
        for (i = 0; i < lo->nslot; i++) {
            struct nbd_slot * sloti = & lo->slots[i];
            if (sloti->file)
                lo->aslot++;
        }
        NBD_DEBUG (3, "decremented active socket count to %d\n",
                lo->aslot);

        if (lo->aslot <= 0) {
           lo->flags &= ~NBD_ENABLED;
           NBD_DEBUG (3, "disabled device nd%c\n", lo->nbd + 'a');
        } else {
           lo->flags |= NBD_ENABLED;
           NBD_DEBUG (3, "enabled device nd%c\n", lo->nbd + 'a');
        }

    }

    /* PTB reset lo->islot */

    if (lo->islot == islot) {
        for (i = 0; i++ < lo->nslot;) {
            lo->islot++;
            if (lo->islot >= lo->nslot)
                lo->islot = 0;
            if (lo->slots[lo->islot].file)
                break;
        }
        NBD_DEBUG (3, "reset current socket slot to %d\n", lo->islot);
    }

    init_MUTEX(& slot->lock); /* READER (lie) to let writes proceed */

    NBD_DEBUG (3, "reset semaphore %lx\n",
            (unsigned long) & slot->lock);

    lo->harderror = 0;
    NBD_DEBUG (3, "reset error to 0\n");

    NBD_SEMAPHORE_UP (&lo->queue_lock);         /* PTB */
        
    /* PTB don't clear whole device queue as we might still be open */

    NBD_DEBUG (3, "exited OK\n");
    return 0;
}

/*
 * PTB - check slots for old requests and roll them back. 
 */
static void
nbd_rollback_old(struct nbd_slot *slot){

    int islot = slot->i;

    NBD_DEBUG (3, "entered\n");

    if (slot->req_age > 0 && slot->req_age < jiffies - NBD_AGE_REQ * HZ )  {
       NBD_DEBUG (1, "try rollback req on slot %d\n", islot);
       nbd_rollback(slot);
    }
    
    NBD_DEBUG (3, "exited OK\n");
}

/*
 * PTB - check all slots for old requests and roll them back. 
 */
static void
nbd_rollback_old_all(struct nbd_device *lo) {

    int islot;
    NBD_DEBUG (3, "entered\n");

    for (islot = 0; islot < lo->nslot; islot++){
      struct nbd_slot * slot = & lo->slots[islot];
      if (slot->req_age > 0 && slot->req_age < jiffies - NBD_AGE_REQ * HZ )  {
           NBD_DEBUG (1, "try rollback req on slot %d\n",
                   islot);
           nbd_rollback(slot);
      }
    }

    NBD_DEBUG (3, "exited OK\n");
}


/*
 * PTB - register a socket to a slot.
 *     - Return 0 for success and -ve for failure.
 */
static int
nbd_set_sock (struct nbd_slot * slot, int arg)
{

    struct file *file      = NULL;
    struct inode *inode    = NULL;
    struct nbd_device *lo  = slot->lo;
    int islot = slot->i;

    NBD_DEBUG (3, "entered with arg %d on slot %d\n"
               ,arg, islot);

    /* this is the file for the fd of the socket */
    file = fget(arg); 
    NBD_DEBUG (3, "fget returned %x\n", (unsigned) file);
    if (!file) {
        NBD_ALERT ("NULL file?!?\n");
        NBD_DEBUG (3, "exited BADF\n");
        return -EBADF;
    }
    NBD_DEBUG (3, "I got file %x\n", (unsigned) file);
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
    inode = file->f_dentry->d_inode;    /* PTB */
#else
    inode = file->f_inode;      /* PTB */
#endif
    /* N.B. Should verify that it's a socket */
    if (!inode) {
        NBD_ALERT ("NULL inode?!?\n");
        NBD_DEBUG (3, "exited INVAL\n");
        return -EINVAL;
    }
    NBD_DEBUG (3, "got inode %x\n", (unsigned) inode);
    if (!(lo->flags & NBD_INITIALISED))  {
        NBD_ALERT ("Device not initialised yet!\n");
        NBD_DEBUG (3, "exited NODEV\n");
        return -ENODEV;
    }
    if (!(lo->flags & NBD_SIZED)) {
        NBD_ALERT ("Device not sized yet!\n");
        NBD_DEBUG (3, "exited INVAL\n");
        return -EINVAL;
    }
    if (!(lo->flags & NBD_BLKSIZED)) {
        NBD_ALERT ("Device not blksized yet!\n");
        NBD_DEBUG (3, "exited INVAL\n");
        return -EINVAL;
    }
    if (!(lo->flags & NBD_SIGNED)) {
        NBD_ALERT ("Device not signed yet!\n");
        NBD_DEBUG (3, "exited INVAL\n");
        return -EINVAL;
    }
    if (1) {                    /* added by ptb */
        invalidate_inode_pages (inode);
        NBD_DEBUG (3, "invalidated inode %x\n", (unsigned) inode);
    }
    NBD_DEBUG (3, "received request for slot %d with %d active of %d\n",
               islot, lo->aslot, lo->nslot);
    NBD_DEBUG (3, "selected slot %d of %d\n", islot, lo->nslot);

    slot = & lo->slots[islot];

    if (slot->file) {
        int error = -EBUSY;     /* PTB need something better */
        NBD_DEBUG (3, "socket %d is not empty\n", islot);
        NBD_DEBUG (3, "exited with error %d\n", error);
        return error;
    }
    NBD_DEBUG (3, "verified slot %d of %d is available\n",
      islot, lo->nslot);
    #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
    file->f_count++;
    #else
    atomic_inc(& file->f_count);
    #endif
    NBD_DEBUG (3, "incremented file reference count\n");
    NBD_SEMAPHORE_DOWN (&lo->queue_lock);
    slot->file = file;
    NBD_DEBUG (1, "set file reference %d to %ld\n",
           islot, (long) file);
    slot->sock = &inode->u.socket_i;
    NBD_DEBUG (1, "set socket %d reference on inode %x\n",
               islot, (unsigned) inode);

    if (++lo->aslot > 0) {
      lo->flags |= NBD_ENABLED;
      NBD_DEBUG (1, "enabled device nd%c\n", lo->nbd + 'a');
    }
    NBD_DEBUG (3, "increased active socket count to %d\n" ,lo->aslot);

    if (islot >= lo->nslot) {
        lo->nslot = islot + 1;
        NBD_DEBUG (3, "increased socket count to %d\n"
                   ,lo->nslot);
    }
    slot->pid = current->pid;
    init_MUTEX_LOCKED(& slot->lock);   /* no reader */
    NBD_DEBUG (3, "set semaphore %lx\n",
        (unsigned long) & slot->lock);
    lo->harderror = 0;
    NBD_DEBUG (3, "set error to 0\n");
    NBD_SEMAPHORE_UP (&lo->queue_lock);
    NBD_DEBUG (3, "exited OK with socket slot %d\n", islot);
    return 0;
}

/*
 * PTB - return the index i of 2^i + j, 0 <= j < 2^i
 */
static inline unsigned 
log2 (unsigned arg)
{
    unsigned log = 0;
    while ((arg >>= 1) > 0)
        log++;
    return log;
}

/*
 * PTB - set the blksize in bytes of the block device. Return 0 for
 *     - success and -ve for failure.
 */
static int
nbd_set_blksize (struct nbd_device *lo, unsigned int arg)
{
    int nbd = lo->nbd;
    NBD_DEBUG (3, "entered with arg %x (%u) on device %d\n"
               ,arg, arg, nbd);
    if ((arg & 511) || (arg > PAGE_SIZE)) {
        NBD_ERROR ("blksize too big (%u)\n", arg);
        NBD_DEBUG (3, "exited INVAL\n");
        return -EINVAL;
    }
    lo->blksize = nbd_blksizes[nbd << NBD_SHIFT] = arg;
    NBD_DEBUG (1, "set blksize to %d \n", lo->blksize);
    lo->logblksize = log2 (lo->blksize);
    lo->flags |= NBD_BLKSIZED;
    NBD_DEBUG (3, "exited OK\n");
    return 0;
}

/*
 * PTB - set the size in bytes of the block device. Return 0 for
 *     - success and -ve for failure.
 */
static int
nbd_set_size (struct nbd_device *lo, u64 arg)
{
    int nbd = lo->nbd;
    NBD_DEBUG (3, "entered with arg %lx%.8lx  on device %d\n",
               (long)(arg>>32), (unsigned long)(arg&0xffffffff),
               nbd);
    lo->bytesize = nbd_bytesizes[nbd << NBD_SHIFT] = arg;
    NBD_DEBUG (1, "set bytesize to %luKB \n",(unsigned long)(lo->bytesize>>10));
    lo->size = nbd_sizes[nbd << NBD_SHIFT] = arg >> 10;
    NBD_DEBUG (1, "set size (KB) to %d \n", lo->size);

#if DEBUG > 0
    if (debug) {
      if (lo->blockmap) {
        kfree(lo->blockmap);
        lo->blockmap = NULL;
      }
      if (!lo->blockmap) {
        /* PTB two bits per block in the blockmap */
        unsigned blocks = lo->size >> (lo->logblksize - 10);
        lo->blockmap = kmalloc(blocks>>2, GFP_KERNEL);
        if (lo->blockmap)
          memset(lo->blockmap, 0xff, blocks>>2);
      }
    } 
#endif

    lo->flags |= NBD_SIZED;
    NBD_DEBUG (3, "exited OK\n");
    return 0;
}

/*
 * PTB - if we're not signed, accept new sig and return success.
 *     - if we are signed, compare the offer and return success if equal,
 *     - and -ve for failure.
 */
static int
my_nbd_set_sig (struct nbd_slot * slot, int * sig) {
    int err = 0;
    int i = 0;
    int buf[NBD_SIGLEN / sizeof(int)];
    int islot = slot->i;
    struct nbd_device * lo = slot->lo;

    NBD_DEBUG (3, "entered\n");
    if (!access_ok(VERIFY_READ, (char*)sig, NBD_SIGLEN)){
        err = -EINVAL;
        NBD_DEBUG(3, "(%d) exited INVAL\n" , islot);
        return err;
    }
    if (!(lo->flags & NBD_SIGNED)) {
        /* PTB first time grab sig */
        copy_from_user ((char*) lo->signature, (char*) sig, NBD_SIGLEN);
        lo->flags |= NBD_SIGNED;
        slot->pid = current->pid;
        NBD_DEBUG (3, "(%d): set sig and exited OK\n", islot);
        return 0;
    }
    copy_from_user ((char*)buf, (char*)sig, NBD_SIGLEN);

    /* PTB test for equality */

    for (i = 0; i < NBD_SIGLEN / sizeof(int); i++) 
      if (buf[i] != lo->signature[i]) {
        err = -EINVAL;
        NBD_ALERT("(%d): failed sigcheck wth %d\n", islot, err );
        return err;
      }
    slot->pid = current->pid;
    err = 0;
    NBD_DEBUG(3, "(%d): verified sig and exited OK\n", islot);
    return err;
}

/*
 * PTB - register a userspace buffer to a slot. Return 0 for success
 *     - and -ve for failure. Null arg acts as erase.
 */
static int
my_nbd_reg_buf (struct nbd_slot * slot,  char * buffer) {

    int err=0, siz ;
    struct nbd_device * lo = slot->lo;
    int maxblks = buf_sectors >> (lo->logblksize - 9);
    int islot = slot->i;

    NBD_DEBUG (3, "entered\n");

    #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
    if (blk_dev[MAJOR_NR].request_fn != nbd_do_request) { 
        err = -EINVAL;
        NBD_DEBUG(3, "(%d): exited INVAL\n", islot);
        return err;
    }
    #endif

    if (!buffer) {
       slot->flags &= ~NBD_SLOT_BUFFERED;
       NBD_DEBUG(1, "(%d): unregistered buffer %x size %d\n",
       islot, (unsigned) slot->buffer, slot->bufsiz);
       slot->buffer = NULL;
       slot->bufsiz = 0;
       return 0;
    }

    slot->buffer = buffer;
    slot->bufsiz = siz = lo->blksize + sizeof(struct nbd_request);

    /* verify the buffer holds one block */
    if (!access_ok(VERIFY_WRITE, buffer, siz)){
        err = -EINVAL;
        NBD_DEBUG(3, "(%d) exited with %d\n" , islot, err );
        return err;
    }
    /* PTB hope the buffer is as big as it should be - FIXME */
    siz = maxblks * lo->blksize;
    slot->bufsiz = siz;

    /* PTB let the device bufsiz be min of registered nonzero bufsizes */
    if (!lo->bufsiz)
       lo->bufsiz = siz;
    else
       if (lo->bufsiz > siz)
           lo->bufsiz = siz; 

    slot->flags |= NBD_SLOT_BUFFERED;
    NBD_DEBUG(1, "(%d): registered buffer %x size %d\n",
       islot, (unsigned) buffer, siz);
    NBD_DEBUG (3, "exited OK\n");
    return 0;
}

static void nbd_enable(struct nbd_device *lo);
static struct timer_list reenable_timer = {
    NULL,
    0,
    0,
    0,
    (void (*)(unsigned long))nbd_enable,
};

/*
 * PTB - set the enabled flag on a slot
 */
static void nbd_enable(struct nbd_device *lo) {
   if (!(lo->flags & NBD_INITIALISED)) 
     return;
   lo->flags |= NBD_ENABLED;
}

/*
 * PTB - drains full device queue and zeros mod count.
 */
static int 
nbd_soft_reset (struct nbd_device *lo)
{
    int j = 0;
    NBD_DEBUG (3, "entered\n");
    if (!(lo->flags & NBD_INITIALISED && lo->nslot > 0))
        return -EINVAL;
    for (; j < lo->nslot; j++) {
        struct nbd_slot * slot = & lo->slots[j];
        NBD_DEBUG (1, "rollback queue on slot %d\n", j);
        nbd_rollback (slot);
    }
    NBD_DEBUG (1, "clear queue\n");
    // NBD_SEMAPHORE_DOWN (&lo->queue_lock);
    nbd_clr_queue (lo);
    // NBD_SEMAPHORE_UP (&lo->queue_lock);
    //lo->flags &= ~NBD_SIGNED;   /* PTB unsign the device */
    invalidate_buffers (MKDEV(NBD_MAJOR,lo->nbd << NBD_SHIFT));
    lo->flags &= ~NBD_ENABLED;
    reenable_timer.data = (unsigned long)lo;
    reenable_timer.expires = jiffies + 5 * HZ;
    add_timer(&reenable_timer);
    NBD_DEBUG (3, "exited OK\n");
    return 0;
}

/*
 * PTB - added a device/module reset for tidyness in face of rampant hacking
 *     - this does a nbd_reset of all devices.
 */
int
nbd_hard_reset ()
{
    int i;
#ifdef MODULE
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
    int old_count = GET_USE_COUNT(&__this_module);
#else
    int old_count = mod_use_count_;
#endif
#endif
    int err = 0;
    NBD_DEBUG (3, "entered\n");

    for (i = 0; i < MAX_NBD; i++)
            nbd_soft_reset (&nbd_dev[i]);
#ifdef MODULE
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
    while (MOD_IN_USE) 
           MOD_DEC_USE_COUNT;
#else
    mod_use_count_ = 0 ;
#endif
#endif

    NBD_DEBUG (1, "altered usage count to 0 from %d\n",
                   old_count);

    NBD_DEBUG (3, "exited\n");
    return err;
}

/*
 * PTB - generic ioctl handling
 */
static int
nbd_ioctl (struct inode *inode, struct file *file,
           unsigned int cmd, unsigned long arg)
{
    struct nbd_device *lo = 0;
    int minor = -1;  /* PTB minor on which we got the ioctl */
    int islot = -1;
    int nbd = -1;   /* the count for the device group */
    struct nbd_slot * slot = 0;

    NBD_DEBUG (3, "entered with arg %lu (%lx)\n", arg, arg);

    if (!suser ()) {
        NBD_ERROR ("caller must be root.\n");
        NBD_DEBUG (3, "exited PERM\n");
        return -EPERM;
    }
    if (!inode) {
        NBD_ERROR ("given bad inode.\n");
        NBD_DEBUG (3, "exited INVAL\n");
        return -EINVAL;
    }
    if (MAJOR (inode->i_rdev) != MAJOR_NR) {
        NBD_ERROR ("pseudo-major != %d\n", MAJOR_NR);
        NBD_DEBUG (3, "exited NODEV\n");
        return -ENODEV;
    }
    minor = MINOR (inode->i_rdev);
    nbd = minor >> NBD_SHIFT;
    if (nbd >= MAX_NBD) {
        NBD_ERROR ("tried to open too many devices, %d\n", minor);
        NBD_DEBUG (3, "exited NODEV\n");
        return -ENODEV;
    }
    lo = & nbd_dev[nbd];
    lo->harderror = 0;
    islot = minor % NBD_MAXCONN - 1;
    slot = &lo->slots[islot];

    switch (cmd) {
          int err;

      case NBD_CLEAR_SOCK:
          NBD_DEBUG (1, "NBD_CLEAR_SOCK %ld\n", arg);
          err = nbd_clr_sock (slot);
          if (lo->aslot <= 0) { /* PTB - probably incorrect, not sock oriented */
              invalidate_buffers (MKDEV(NBD_MAJOR, nbd<<NBD_SHIFT));
              NBD_DEBUG (3, "invalidated buffers on device %c\n", 'a'+nbd);
          }

          NBD_DEBUG (1, "exited with %d\n", err);
          return err;

      case NBD_SET_SOCK:
          NBD_DEBUG (3, "NBD_SET_SOCK %ld\n", arg);
          err = nbd_set_sock (slot, arg);
          NBD_DEBUG (3, "exited with %d\n", err);
          return err;

      case NBD_GET_BLKSIZE:
          NBD_DEBUG (3, "NBD_GET_BLKSIZE %lx\n", arg);
          if (!access_ok(VERIFY_WRITE, (long *) arg, sizeof(long))) {
              err = -EINVAL;
              NBD_DEBUG (3, "exited INVAL\n");
              return err;
          }
          if (!(lo->flags & NBD_BLKSIZED)) {
              NBD_DEBUG (1, "GET_BLKGETSIZE exited INVAL\n");
              err = -EINVAL;
              return err;
          }
          put_user (lo->blksize, (long *) arg);
          err = 0;
          NBD_DEBUG (3, "exited OK\n");
          return err;

      case NBD_SET_BLKSIZE:
          NBD_DEBUG (3, "NBD_SET_BLKSIZE %ld\n", arg);
          err = nbd_set_blksize (lo, arg);
          NBD_DEBUG (3, "exited with %d\n", err);
          return err;

      case NBD_SET_SIZE:
          NBD_DEBUG (3, "NBD_SET_SIZE %ld\n", arg);
          err = nbd_set_size (lo, (u64)arg);
          NBD_DEBUG (3, "exited with %d\n", err);
          return err;

      case NBD_SET_SECTORS:
          NBD_DEBUG (3, "NBD_SET_SECTORS %ld\n", arg);
          err = nbd_set_size (lo, ((u64)arg) << 9);
          NBD_DEBUG (3, "exited with %d\n", err);
          return err;

      case MY_NBD_REG_BUF: /* PTB register your buffer per socket here */
          NBD_DEBUG (3, "MY_NBD_REG_BUF %lx\n", arg);
          if (!arg) {
              /* PTB serves as existence check for this ioctl */
              NBD_DEBUG (3, "exited OK\n");
              return 0;
          }
          err = my_nbd_reg_buf (slot, (char *) arg);
          NBD_DEBUG(3, "(%d): exited with %d\n" , islot, err );
          return err;

      case MY_NBD_SET_SIG: 
          NBD_DEBUG (3, "MY_NBD_SET_SIG %lx\n", arg);
          err = my_nbd_set_sig (slot, (int *) arg);
          NBD_DEBUG(3, "(%d): exited with %d\n" , islot, err );
          return err;

      case MY_NBD_GET_REQ:
          NBD_DEBUG (3, "MY_NBD_GET_REQ %lx\n", arg);
          if (!slot) {
              err = -EINVAL;
              NBD_DEBUG(3, "exited INVAL\n");
              return err;
          }
          err = nbd_get_req (slot);
          if (err < 0) {
            /* AMARIN Something missing here? */
          }
          NBD_DEBUG(3, "exited with %d\n" , err );
          return err;

      case MY_NBD_CLR_REQ:
          NBD_DEBUG (3, "MY_NBD_CLR_REQ %ld\n", arg);
          nbd_rollback(slot);
          err = 0;
          NBD_DEBUG (3, "exited with %d\n", err);
          return err;

      case MY_NBD_ERR_REQ:
          NBD_DEBUG (3, "MY_NBD_ERR_REQ %ld\n", arg);
          nbd_error(slot);
          err = 0;
          NBD_DEBUG (3, "exited with %d\n", err);
          return err;

      case MY_NBD_SYNC:
          NBD_DEBUG (3, "MY_NBD_SYNC %ld\n", arg);
          if ((long)arg >= 0)
            nbd_rollback_old (slot);
          else
            nbd_rollback_old_all (lo);
          err = 0;
          NBD_DEBUG (3, "exited with %d\n", err);
          return err;

      case MY_NBD_ACK:
          NBD_DEBUG (3, "MY_NBD_ACK %lx\n", arg);
          if (!slot) {
              err = -EINVAL;
              NBD_DEBUG(3, "exited INVAL\n");
              return err;
          }
          err = nbd_ack (slot);
          NBD_DEBUG(3, "exited with %d\n" , err );
          return err;

          /* let this be compiled in always - it's useful. PTB */
      case NBD_PRINT_DEBUG:
          NBD_DEBUG (3, "NBD_PRINT_DEBUG");
          NBD_INFO ("device %d: head = %x, tail = %x, in = %d, out = %d\n"
                    ,minor, (int) NBD_LIST_HEAD(lo), (int) NBD_LIST_TAIL(lo),
                    lo->requests_in, lo->requests_out);
          err = 0;
          NBD_DEBUG (3, "exited OK\n");
          return err;


      case BLKFLSBUF:
          NBD_DEBUG (1, "BLKFLSBUF %lx\n", arg);
          if (!suser()) {
              err =  -EACCES;
              NBD_DEBUG (1, "BLKFLSBUF exited ACCES\n");
              return err;
          }
          if (lo->inode)
             fsync_dev(lo->inode->i_rdev);
          else
             fsync_dev(MKDEV(NBD_MAJOR, nbd<<NBD_SHIFT));
          if (lo->inode)
             invalidate_buffers(lo->inode->i_rdev);
          else
             invalidate_buffers(MKDEV(NBD_MAJOR, nbd<<NBD_SHIFT));
          err = 0;
          NBD_DEBUG (3, "exited OK\n");
          return err;


      case BLKRAGET:
          NBD_DEBUG (1, "BLKRAGET %lx\n", arg);
          if (!access_ok(VERIFY_WRITE, (long *)arg, sizeof(long))) {
              err = -EINVAL;
              NBD_DEBUG (1, "BLKFLSBUF exited FAIL\n");
              return err;
          }
          put_user(read_ahead[MAJOR_NR], (long *) arg);
          err = 0;
          NBD_DEBUG (3, "exited OK\n");
          return err;

      case BLKRASET:
          NBD_DEBUG (1, "BLKRASET %lx\n", arg);
          if (!suser()) {
              err =  -EACCES;
              NBD_DEBUG (1, "BLKRASET exited ACCES\n");
              return err;
          }
          if (arg > 0xff) {
              err =  -EINVAL;
              NBD_DEBUG (1, "BLKRASET exited INVAL\n");
              return err;
          }
          rahead = read_ahead[MAJOR_NR] = arg;
          err = 0;
          NBD_DEBUG (3, "exited OK\n");
          return err;
          
      case HDIO_GETGEO:
          NBD_DEBUG (1, "HDIO_GETGEO %lx\n", arg);
          if (!arg) {
              err =  -EINVAL;
              NBD_DEBUG (1, "HDIO_GETGEO exited INVAL\n");
              return err;
          }
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
          {
            struct hd_geometry * geo = (struct hd_geometry *) arg;
            int sectors = nbd_sizes[nbd << NBD_SHIFT] * 2;
            sectors &= ~0x3f;
            err = 0;
            if((err = put_user(sectors >> 6, &geo->cylinders), err<0)
            || (err = put_user(           4, &geo->heads),     err<0)
            || (err = put_user(          16, &geo->sectors),   err<0)
            || (err = put_user(           4, &geo->start),     err<0)) {
              NBD_DEBUG (1, "HDIO_GETGEO exited FAIL\n");
              return err;
            }
          }
#else
          err = verify_area(VERIFY_WRITE, (char*)arg, sizeof(struct hd_geometry));
          if (err) {
              NBD_DEBUG (1, "HDIO_GETGEO exited FAIL\n");
              return err;
          } else {
            struct hd_geometry * geo = (struct hd_geometry *) arg;
            int sectors = nbd_sizes[nbd << NBD_SHIFT] * 2;
            sectors &= ~0x3f;
            put_user(sectors >> 6, &geo->cylinders);
            put_user(           4, &geo->heads);
            put_user(          16, &geo->sectors);
            put_user(           4, &geo->start);
          }
#endif
          err = 0;
          NBD_DEBUG (3, "exited OK\n");
          return err;

      case BLKGETSIZE:          /* PTB 132 */
          /* PTB return nr sectors */
          NBD_DEBUG (1, "BLKGETSIZE %lx\n", arg);
          if (!access_ok(VERIFY_WRITE, (long *) arg, sizeof(long))) {
              err = -EINVAL;
              NBD_DEBUG (3, "invalid user area %lu\n", arg);
              NBD_DEBUG (1, "BLKGETSIZE exited INVAL\n");
              return err;
          }
          if (!(lo->flags & NBD_SIZED)) {
              NBD_DEBUG (1, "BLKGETSIZE exited NODEV\n");
              err = -ENODEV;
              return err;
          }
          put_user ((unsigned long) (lo->size << 1), (unsigned long *)arg);
          // PTB this check is silly here and seems to trigger!
          if (lo->size != 0 && (u32)(lo->bytesize >> 10) != lo->size) {
              NBD_ALERT ("bytes %luKB mismatch with KB %u in BLKGETSIZE\n"
                     , (unsigned long)(lo->bytesize>>10), lo->size);
          } else {
              NBD_DEBUG (1, "bytes %luKB match with KB %u in BLKGETSIZE\n"
                     , (unsigned long)(lo->bytesize>>10), lo->size);
          }
          err = 0;
          NBD_DEBUG (1, "exited OK\n");
          return err;

      case NBD_HARD_RESET:     /* PTB - debugging */
          NBD_DEBUG (3, "NBD_HARD_RESET\n");
          err = nbd_hard_reset();
          NBD_DEBUG (3, "exited OK\n");
          return err;
          
      case NBD_RESET:          /* PTB - debugging */
          NBD_DEBUG (3, "NBD_RESET\n");
          err = nbd_soft_reset(lo);
          NBD_DEBUG (3, "exited OK\n");
          return err;

      case NBD_SET_RO:        /* PTB - change ro status */
          NBD_DEBUG (3, "NBD_SET_RO %ld\n", arg);
          if (arg)
            lo->flags |= NBD_READ_ONLY;
          else
            lo->flags &= ~NBD_READ_ONLY;
          err = 0;
          NBD_DEBUG (3, "exited OK\n");
          return err;

      case NBD_DEC_USE_COUNT: /* AMARIN */
          NBD_DEBUG(3,"nbd_dec_use_count enter, MOD_IN_USE=%s\n" ,
               MOD_IN_USE?  "yes" : "no");
          while (MOD_IN_USE > 0){
            MOD_DEC_USE_COUNT;
            NBD_DEBUG(3,"Decrementing mod_use_count");
          }
          invalidate_buffers (MKDEV(NBD_MAJOR, nbd << NBD_SHIFT));
          NBD_DEBUG(3,"nbd_dec_use_count exited, MOD_IN_USE=%s\n" ,
               MOD_IN_USE?  "yes" : "no");
          return 0;

    }
    NBD_DEBUG (3, "exited INVAL\n");
    return -EINVAL;
}

/*
 * PTB - release the device
 */
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
static int
#else
static void
#endif
nbd_release (struct inode *inode, struct file *file)
{
    struct nbd_device *lo;
    int dev;
    int nbd;

    NBD_DEBUG (2, "entered\n");

    if (!inode) {
        NBD_ALERT ("null inode.\n");
        NBD_DEBUG (3, "exited NODEV\n");
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
        return -ENODEV;
#else
        return;
#endif
    }
    dev = MINOR (inode->i_rdev);
    nbd = dev >> NBD_SHIFT;

    if (nbd >= MAX_NBD ) {
        NBD_ALERT ("too many open devices.\n");
        NBD_DEBUG (3, "exited NODEV\n");
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
        return -ENODEV;
#else
        return;
#endif
    }

    lo = &nbd_dev[nbd];

    if (lo->inode)
         fsync_dev (lo->inode->i_rdev);
    else
         fsync_dev (MKDEV(NBD_MAJOR, nbd << NBD_SHIFT));

    if (lo->refcnt <= 0) {
        NBD_ALERT ("refcount(%d) <= 0\n", lo->refcnt);

#if 0
        if (lo->inode)
             invalidate_buffers (lo->inode->i_rdev);
        else
             invalidate_buffers (MKDEV(NBD_MAJOR, nbd<<NBD_SHIFT));
        NBD_DEBUG (2, "invalidated buffers\n");
#endif

    }
    if (!NBD_LIST_EMPTY(lo)) {
        NBD_ALERT ("Some requests are waiting queued -> can not turn off.\n");
        NBD_DEBUG (3, "exited BUSY\n");
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
        return -EBUSY;
#else
        return;
#endif
    }
    if (lo->requests_req > 0) {
        NBD_ALERT ("Some requests are counted in progress -> can not turn off.\n");
        NBD_DEBUG (3, "exited BUSY\n");
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
        return -EBUSY;
#else
        return;
#endif
    }

    /* POSSIBLE change socket here PTB */

    lo->refcnt--;
    NBD_DEBUG (3, "decrement device %c reference count.\n",
        'a' + nbd);
    if (MOD_IN_USE)
        MOD_DEC_USE_COUNT;

    if (lo->refcnt <= 0 || ! MOD_IN_USE) {
       if (lo->inode)
         invalidate_buffers (lo->inode->i_rdev);     
       else
         invalidate_buffers (MKDEV(NBD_MAJOR, nbd << NBD_SHIFT));
       NBD_DEBUG (2, "invalidated buffers\n");

       lo->bufsiz = 0;
    }

    NBD_DEBUG (2, "exited OK with %d requests in queue\n",
               lo->count);
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
    return 0;
#else
        return;
#endif
}

#if 0
/*
 * PTB - Do a reverse-flush of the device to the queue, either of our
 *     - slot if called from a client minor, or of the whole device if called
 *     - from the whole device minor.
 */
int 
nbd_fsync (struct inode *inode, struct file *file)
{
    int dev = -1;
    struct nbd_device *lo = 0;
    int err = 0;
    int nbd = -1;
    int islot = -1;

    NBD_DEBUG (3, "entered\n");

    if (1 && !inode && file) {
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
        inode = file->f_dentry->d_inode;
#else
        inode = file->f_inode;
#endif
    }
    if (!inode) {
        NBD_ERROR ("nbd_fsync - null inode.\n");
        NBD_DEBUG (3, "exited INVAL\n");
        return -EINVAL;
    }

    NBD_DEBUG (3, "have inode %x\n", (unsigned) inode);

    dev = MINOR (inode->i_rdev);
    nbd = dev >> NBD_SHIFT;
    islot = dev - (nbd << NBD_SHIFT) - 1;

    if (nbd >= MAX_NBD) {
        NBD_ERROR ("nbd_fsync - too many (%d) devices open\n", nbd);
        NBD_DEBUG (3, "exited NODEV\n");
        return -ENODEV;
    }

    NBD_DEBUG (3, "have minor device %d\n", dev);

    lo = &nbd_dev[nbd];
    slot = & lo->slots[islot];

    if (islot < 0)
      nbd_rollback_old_all (lo);
    else
      nbd_rollback_old_all (slot);

    NBD_DEBUG (3, "exited\n");

    return err = 0;
}
#endif

#if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
static struct file_operations nbd_fops =
{
        NULL,                       /* lseek - default */
        block_read,                 /* read - general block-dev read */
        block_write,                /* write - general block-dev write */
        NULL,                       /* readdir - bad */
        NULL,                       /* poll/select */
        nbd_ioctl,                  /* ioctl */
        NULL,                       /* mmap */
        nbd_open,                   /* open */
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
        NULL,                       /* flush */
#endif
        nbd_release,                /* release */
        NULL,
};
#else
static struct block_device_operations nbd_blkops =
{
        open: nbd_open,
        release: nbd_release,
        ioctl: nbd_ioctl,
        check_media_change: NULL,
        revalidate: NULL,
};
#endif


/*
 * And here should be modules and kernel interface 
 *  (Just smiley confuses emacs :-)
 */

int 
#if LINUX_VERSION_CODE < LINUX_VERSION_2_1_0
nbd_get_info (char *buf, char **start, off_t offset, int len, int unused)
#else
nbd_read_proc (char *buf, char **start, off_t offset, int len, int *eof, void *data)
#endif
{

#ifndef MIN
  #define MIN(x,y) ((x)<(y)?(x):(y))
#endif

    const int limit = MIN(PAGE_SIZE,len) - 80 ;
    static int i;
    struct nbd_device *lo;
    static int last;
    static int last_despatched[MAX_NBD];
    extern volatile unsigned long jiffies; 
    static int stuck;
    static int alive;
    static void *next_label;
    static char *next_label_name;
    static int total;

    NBD_DEBUG (2, "entered len=%d, offset=%d, next_label=%s\n",
       len, (int)offset, next_label_name?:"null");

    if (offset > 0 && !next_label) {
    #if LINUX_VERSION_CODE >= LINUX_VERSION_2_1_0
      *eof = 1;
    #endif
      *start = buf;
      NBD_DEBUG (2, "exited OK after 0 bytes total %d\n", total);
      return 0;
    }

    if (offset <= 0) {
      // PTB do static inits first time through
      last = -1;
      i = 0;
      next_label = NULL;
      next_label_name = NULL;
      stuck = 0;
      alive = 0;
      total = 0;
    }

    // PTB  start this bytecount
    len = 0;

    NBD_DEBUG (2, "device=%d, last device=%d, next label %s\n",
          i, last, next_label_name?:"null");

#if LINUX_VERSION_CODE < LINUX_VERSION_2_1_0
#define NBD_PROC_LABEL(n) \
        next_label = &&label_##n; \
        next_label_name = "label_" #n; \
        if (len > limit) { \
            total += len; \
            NBD_DEBUG (2, "short return in show device %d after %d bytes " \
            "written to total %d bytes " \
            "with next label %s\n", i, (int)len, total, next_label_name); \
            return len;\
        } \
        NBD_DEBUG (2, "passed %s in show device %d after %d bytes\n", \
          next_label_name, i, (int)len); \
        label_##n:
#else
#define NBD_PROC_LABEL(n) \
        next_label = &&label_##n; \
        next_label_name = "label_" #n; \
        if (len > limit) { \
            *start = (char *) len; \
            total += len; \
            NBD_DEBUG (2, "short return in show device %d after %d bytes " \
            "written to total %d bytes " \
            "with next label %s\n", i, (int)len, total, next_label_name); \
            return len;\
        } \
        NBD_DEBUG (2, "passed %s in show device %d after %d bytes\n", \
          next_label_name, i, (int)len); \
        label_##n:
#endif

    for (/* static init */; i < MAX_NBD; i++) {

        lo = &nbd_dev[i];
        if (lo->nslot <= 0) {
            next_label = NULL;
            continue;
        }

        // PTB computed goto next not-done
        if (next_label) {
          void * label = next_label;
          NBD_DEBUG (2, "did computed goto %s in show device %d on reentry\n",
                next_label_name, i);
          next_label   = NULL;
          next_label_name = NULL;
          len = 0;
          goto *label;
        }

        NBD_PROC_LABEL(1);

        if (last == i - 2)
            len += sprintf (buf + len, "Device %c:\tClosed\n", 'a' + i - 1);
        if (last < i - 2)
            len += sprintf (buf + len, "Device %c-%c:\tClosed\n",
                'a' + last + 1, 'a' + i - 1);

        NBD_PROC_LABEL(2);

        len += sprintf (buf + len, "Device %c:\tOpen\n", 'a' + i);
        alive++;

        NBD_PROC_LABEL(3);

        len += sprintf (buf + len,
             "[%c] State:\t%s, %s, %s,%s %s, last error %d\n", 'a' + i,
              lo->flags & NBD_INITIALISED ? "initialized" : "uninitialized",
              lo->flags & NBD_WRITE_NOCHK ? "noverify" : "verify",
              lo->flags & NBD_READ_ONLY ? "ro" : "rw",
              merge_requests?" merge requests,":"",
              lo->flags & NBD_ENABLED ? "enabled" : "disabled",
              lo->harderror);

        NBD_PROC_LABEL(4);

        do {                    // PTB begin long do once block
            int count = 0;

            #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
              struct request * pos;
            #else
              struct list_head * pos;
            #endif

            NBD_SEMAPHORE_DOWN (&lo->queue_lock);

            NBD_DEBUG (3, "counting queue length %d\n", lo->count);
            NBD_LIST_FOR_EACH(pos,lo) {
              struct request * req = NBD_LIST_ENTRY(pos); /* PTB comp. */
              NBD_DEBUG (3, "req %d on major queue is %x\n", count,
                    (unsigned)req);
              count++;
            }
            NBD_DEBUG (3, "counted queue length %d\n", count);

            NBD_SEMAPHORE_UP (&lo->queue_lock);

            NBD_DEBUG (3, "queue length %d\n", count);
            len += sprintf (buf + len,
               "[%c] Queued:\t+%d curr reqs/+%d real reqs/+%d max reqs\n",
                        'a' + i, lo->count, count, lo->maxq);
        PARANOIA_BEGIN;
            if (count != lo->count) {
                /* PTB last ditch! HACK HACK HACK FIXME */
                NBD_DEBUG (1, ":"
                  " altered queued count to %d from %d\n", count, lo->count);
                lo->count = count;
            }
        PARANOIA_END;
        } while(0);                   // PTB end long do once block

        NBD_PROC_LABEL(5);
           
        len += sprintf (buf + len, "[%c] Buffersize:\t%d\t(sectors=%d)\n",
                        'a' + i, lo->bufsiz, lo->bufsiz >> 9);
        len += sprintf (buf + len, "[%c] Blocksize:\t%d\t(log=%d)\n",
                        'a' + i, lo->blksize, lo->logblksize);
        len += sprintf (buf + len, "[%c] Size:\t%luKB\n",
                        'a' + i, (unsigned long)(lo->bytesize >> 10));
        len += sprintf (buf + len, "[%c] Blocks:\t%u\n",
                        'a' + i, (lo->size << 1) / (lo->blksize >> 9));

        NBD_PROC_LABEL(6);

        len += sprintf (buf + len,
                        "[%c] Sockets:\t%d",
                        'a' + i, lo->nslot);

        NBD_PROC_LABEL(7);

        do {                           // PTB begin short do once block
            int j;
            for (j = 0; j < lo->nslot; j++) {
                struct nbd_slot * slotj = & lo->slots[j];
                if (j != lo->islot )
                  len += sprintf (buf + len, "\t(%s)", slotj->file ? "+" : "-");
                else
                  len += sprintf (buf + len, "\t(%s)", slotj->file ? "*" : ".");
            }
        } while(0);                   // PTB end short do once block

        len += sprintf (buf + len, "\n");

        NBD_PROC_LABEL(8);

        /* PTB count blocks queued on general queue */
        do {
            int pending_blks = 0; /* PTB  not reached the slots yet */

          #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
            struct request *pos;
          #else
            struct list_head * pos;
          #endif

            struct request *req;
            int count = 0;

            NBD_SEMAPHORE_DOWN (&lo->queue_lock);

            NBD_DEBUG (3, "counting queued blocks\n");
            NBD_LIST_FOR_EACH(pos,lo) {
              req = NBD_LIST_ENTRY(pos); /* PTB comp. */
              NBD_DEBUG (3, "req %d on major queue is %x\n", count,
                 (unsigned)req);
              count++;
              pending_blks += nr_blks(lo,req);
            }
            NBD_DEBUG (3, "counted queued blocks %d\n", pending_blks);

            NBD_DEBUG (3, "queued blocks %d\n", pending_blks);
            NBD_SEMAPHORE_UP (&lo->queue_lock);
            len += sprintf (buf + len,
                        "[%c] Requested:\t%d+%d",
                        'a' + i, lo->requests_in - pending_blks, pending_blks);
        } while(0);

        NBD_PROC_LABEL(9);

        do {                           // PTB begin short do once block
            int j;
            for (j = 0; j < lo->nslot; j++) {
                struct nbd_slot * slotj = & lo->slots[j];
                len += sprintf (buf + len, "\t(%d)", slotj->in);
            }
        } while(0);                    // PTB end short do once block

        len += sprintf (buf + len, "\n");
        len += sprintf (buf + len,
                        "[%c] Dispatched:\t%d",
                        'a' + i, lo->requests_out);

        NBD_PROC_LABEL(10);

        do {                           // PTB begin short do once block
            int j;
            for (j = 0; j < lo->nslot; j++) {
                struct nbd_slot * slotj = & lo->slots[j];
                len += sprintf (buf + len, "\t(%d)", slotj->out);
            }
        } while(0);                    // PTB end short do once block

        len += sprintf (buf + len, "\n");
        len += sprintf (buf + len,
                        "[%c] Errored:\t%d",
                        'a' + i, lo->requests_err);

        NBD_PROC_LABEL(11);

        do {                           // PTB begin short do once block
            int j;
            for (j = 0; j < lo->nslot; j++) {
                struct nbd_slot * slotj = & lo->slots[j];
                len += sprintf (buf + len, "\t(%d)", slotj->err);
            }
        } while(0);                    // PTB end short do once block

        len += sprintf (buf + len, "\n");

        NBD_PROC_LABEL(12);

        do {                           // PTB begin long do once block
            int pending_blks = 0; /* PTB  not reached the slots yet */
            NBD_SEMAPHORE_DOWN (&lo->queue_lock);

            do {                       // PTB begin short do once block
            #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
              struct request * pos;
            #else
              struct list_head * pos;
            #endif

              int count = 0;
              struct request *req;

              NBD_LIST_FOR_EACH(pos,lo) {
                req = NBD_LIST_ENTRY(pos); /* PTB comp. */
                NBD_DEBUG (3, "req %d on major queue is %x\n", count,(unsigned)req);
                count++;
                pending_blks += nr_blks(lo,req);
              }
            } while(0);                // PTB end short do once block

            NBD_SEMAPHORE_UP (&lo->queue_lock);
            len += sprintf (buf + len,
                        "[%c] Pending:\t%d+%d",
                        'a' + i, lo->requests_req, pending_blks);
        } while(0);                    // PTB end long do once block

        NBD_PROC_LABEL(13);

        do {                           // PTB begin short do once block
            int j;
            for (j = 0; j < lo->nslot; j++) {
                struct nbd_slot * slotj = & lo->slots[j];

              #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
                struct request * pos;
              #else
                struct list_head * pos;
              #endif

                int count = 0;

                NBD_LIST_FOR_EACH(pos,slotj) {
                  struct request * req = NBD_LIST_ENTRY(pos);
                  NBD_DEBUG (3, "req %d on slot %d is %x\n", count,slotj->i,(unsigned)req);
                  count++;
                  len += sprintf (buf + len, "\t(%lu)", nr_blks(lo,req));
                }

            }
        } while(0);                    // PTB end short do once block
        
        len += sprintf (buf + len, "\n");
        if (lo->requests_out == last_despatched[i] && lo->requests_out > 0)
            stuck++;
        last_despatched[i] = lo->requests_out;

        NBD_PROC_LABEL(14);

        len += sprintf (buf + len, "[%c] Kthreads:\t%d", 'a' + i, lo->kthreads);
        len += sprintf (buf + len, "\t(%d waiting/%d running/%d max)\n",
                        lo->kwaiters, lo->kthreads - lo->kwaiters, lo->kmax);

        NBD_PROC_LABEL(15);

        len += sprintf (buf + len, "[%c] Cthreads:\t%d", 'a' + i, lo->cthreads);

        NBD_PROC_LABEL(16);

        do {
            int j;
            for (j = 0; j < lo->nslot; j++) {
                struct nbd_slot * slotj = & lo->slots[j];
                int state = ((slotj->flags & NBD_SLOT_RUNNING)?1:0)
                          + ((slotj->flags & NBD_SLOT_WAITING)?2:0);
                char *desc = "?";
                switch (state) {
                    case 0: desc = "-"; break; /* PTB not in */
                    case 1: desc = "*"; break; /* PTB in and not waiting */
                    case 2: desc = "?"; break; /* PTB impossible */
                    case 3: desc = "+"; break; /* PTB in and waiting */
                }
                len += sprintf (buf + len, "\t(%s)", desc);
            }
        } while(0);

        len += sprintf (buf + len, "\n");
        last = i;
        len += sprintf (buf + len, "[%c] Cpids:\t%d", 'a' + i, lo->cthreads);

        NBD_PROC_LABEL(17);

        do {
            int j;
            for (j = 0; j < lo->nslot; j++) {
                struct nbd_slot * slotj = & lo->slots[j];
                len += sprintf (buf + len, "\t(%u)", slotj->pid);
            }
            len += sprintf (buf + len, "\n");
        } while(0);

        NBD_PROC_LABEL(18);

#if DEBUG > 0

        if (lo->blockmap) {
          static int j;
          static int blocks;
          
          if (blocks <= 0) {
            // first time
            blocks = lo->size >> (lo->logblksize - 10) ;
            j = 0;
          }

          len += sprintf (buf + len, "[%c] blkmap:\n" ,'a' + i);

          while (j < blocks) {
            int k;

            NBD_PROC_LABEL(18_5);

            len += sprintf (buf + len, "[%c] 0x%x:\t", 'a' + i, j*lo->blksize);
            for (k=0;k<64;k++) {
              int offset = j >> 2;
              int shift = (j & 3) << 1;
               if (j >= blocks)
                 break;
               switch ((lo->blockmap[offset] >> shift) & 3) {
                 case READ:
                   len += sprintf (buf + len, "r");
                   break;
                 case READ|2:
                   len += sprintf (buf + len, "R");
                   break;
                 case WRITE:
                   len += sprintf (buf + len, "w");
                   break;
                 default:
                   len += sprintf (buf + len, ".");
                   break;
               }
               j++; 
            }
            len += sprintf (buf + len, "\n");
          }
          j = 0;
        }

#endif

        NBD_PROC_LABEL(19);
    }


    NBD_PROC_LABEL(20);

    if (last == i - 2)
        len += sprintf (buf + len, "Device %c:\tClosed\n", 'a' + i - 1);

    if (last < i - 2) 
        len += sprintf (buf + len, "Device %c-%c:\tClosed\n",
            'a' + last + 1, 'a' + i - 1);

    NBD_PROC_LABEL(21);
    
    // PTB re-init vital statics for next time 
    next_label = NULL;
    next_label_name = NULL;

  #if LINUX_VERSION_CODE >= LINUX_VERSION_2_1_0
    *eof = 1;
  #endif
    *start = buf;
    total += len;

    NBD_DEBUG (2, "exited OK after %d bytes total %d\n", len, total);
    return len;
}

/*  
 * PTB - write a 0 with echo -n 0 to /proc/nbdinfo to do a hard reset.
 */
static
int
nbd_write_proc(struct file *file, const char *buffer, unsigned long count, void *data) {

    NBD_DEBUG (2, "entered with count %lu\n", count);
    switch (count) {
     case 1:
       switch (*buffer - '0') {
          int i;
          case 1: nbd_hard_reset();
                  break;
          case 0: 
                  for (i = 0; i < MAX_NBD; i++)
                          nbd_soft_reset (&nbd_dev[i]);
                  break;
       }
       break;
     default:
         if(!strncmp(buffer,"merge_requests=0",16))
           merge_requests = 0;
         else if(!strncmp(buffer,"merge_requests=1",16))
           merge_requests = 1;
         else if(!strncmp(buffer,"debug=1",7)) {
           int i;
           debug = 1;
           for (i = 0; i < MAX_NBD; i++) {
             struct nbd_device * lo = & nbd_dev[i];
             if (lo->nslot > 0 && !lo->blockmap) {
               unsigned blocks = lo->size >> (lo->logblksize - 10);
               lo->blockmap = kmalloc(blocks>>2, GFP_KERNEL);
               if (lo->blockmap)
                 memset(lo->blockmap, 0xff, blocks>>2);
             }
           }
         } else if(!strncmp(buffer,"debug=0",7)) {
           int i;
           debug = 0;
           for (i = 0; i < MAX_NBD ; i++) {
             struct nbd_device * lo = & nbd_dev[i];
             if (lo->blockmap) {
                kfree(lo->blockmap);
                lo->blockmap = 0;
             }
           }
         }
       break;
    }
    NBD_DEBUG (2, "exited\n");
    return count;
}

#if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
struct proc_dir_entry nbd_proc_entry =
{
    low_ino:0, namelen:7, name:"nbdinfo",
    mode:  S_IFREG|S_IRUGO|S_IWUSR,
    nlink: 1,                      // LNK, DIR or FIL ?
    uid:0, gid:0,                  // uid, gid
    size:  0,                      // size
  #if LINUX_VERSION_CODE < LINUX_VERSION_2_1_0
    get_info:&nbd_get_info,        // get_info 
  #else
    get_info:NULL,
  #endif
    next:  NULL,                   // next entry
    parent:NULL,                   // parent entry
    subdir:NULL,                   // subdir entry
    data:  NULL,                   // void *data
  #if LINUX_VERSION_CODE < LINUX_VERSION_2_1_0
    read_proc:NULL,                // read_proc
  #else
    read_proc:&nbd_read_proc,
  #endif
    write_proc:&nbd_write_proc,    // write_proc
    count: 0,                      // usage count
    deleted:0,                     // delete flag
};
#else
  // PTB this struct is created dynamically
#endif

#ifdef MODULE
#define nbd_init init_module
#if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
    MODULE_AUTHOR("Peter T. Breuer, Andres Marin");
    MODULE_DESCRIPTION("Enhanced Network Block Device");
#endif
#endif

#if LINUX_VERSION_CODE >= LINUX_VERSION_2_3_30
// PTB we steal these from the queue struct at init
static merge_requests_fn *ll_merge_requests_fn;
static merge_request_fn  *ll_front_merge_fn;
static merge_request_fn  *ll_back_merge_fn;
static int
nbd_merge_requests_fn(request_queue_t *q, struct request *req,
    struct request *req2, int max_segments)
{
    if (!merge_requests)
      return 0;
    if (req->nr_sectors + req2->nr_sectors > buf_sectors)
      return 0;
    return ll_merge_requests_fn(q,req,req2,1);
}
static int
nbd_front_merge_fn(request_queue_t *q, struct request *req,
    struct buffer_head *bh, int max_segments)
{
    if (!merge_requests)
      return 0;
    if (req->nr_sectors + (bh->b_size >> 9) > buf_sectors)
      return 0;
    return ll_front_merge_fn(q,req,bh,1);
}
static int
nbd_back_merge_fn(request_queue_t *q, struct request *req,
    struct buffer_head *bh, int max_segments)
{
    if (!merge_requests)
      return 0;
    if (req->nr_sectors + (bh->b_size >> 9) > buf_sectors)
      return 0;
    return ll_back_merge_fn(q,req,bh,1);
}
#endif

int
nbd_init (void)
{
    int i, j;
    int err = 0;
    NBD_DEBUG (3, "entered\n");

    NBD_INFO ("Network Block Device support by pavel@elf.mj.gts.cz\n");
    NBD_INFO ("Network Block Device port to 2.0 by ptb@it.uc3m.es\n");
    NBD_INFO ("Network Block Device trying to put things in user space by amarin@it.uc3m.es\n");
    NBD_INFO ("Enhanced Network Block Device 2.0/2.2/2.4 by ptb@it.uc3m.es\n");
#if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
    if (register_blkdev (MAJOR_NR, "nbd", &nbd_fops)) {
#else
    if (register_blkdev (MAJOR_NR, "nbd", &nbd_blkops)) {
#endif
        NBD_ERROR ("Unable to get major number %d for NBD\n", MAJOR_NR);
        return -EIO;
    }
#ifdef MODULE
    NBD_INFO ("registered device at major %d\n", MAJOR_NR);
#endif
    blksize_size[MAJOR_NR] = nbd_blksizes;      /* blksize in B */
    blk_size    [MAJOR_NR] = nbd_sizes;         /* size in KB */
    blk_init_queue(BLK_DEFAULT_QUEUE(MAJOR_NR), nbd_do_request);
#if LINUX_VERSION_CODE >= LINUX_VERSION_2_3_30
    // PTB disable or at least control merge attempts
    ll_merge_requests_fn = (BLK_DEFAULT_QUEUE(MAJOR_NR))->merge_requests_fn;
    ll_front_merge_fn    = (BLK_DEFAULT_QUEUE(MAJOR_NR))->front_merge_fn;
    ll_back_merge_fn     = (BLK_DEFAULT_QUEUE(MAJOR_NR))->back_merge_fn;
    (BLK_DEFAULT_QUEUE(MAJOR_NR))->merge_requests_fn = &nbd_merge_requests_fn;
    (BLK_DEFAULT_QUEUE(MAJOR_NR))->front_merge_fn    = &nbd_front_merge_fn;
    (BLK_DEFAULT_QUEUE(MAJOR_NR))->back_merge_fn     = &nbd_back_merge_fn;
#endif

    read_ahead  [MAJOR_NR] = rahead;            /* AMARIN/PTB */

    for (i = 0; i < MAX_NBD; i++) {
        struct nbd_device * lo = & nbd_dev[i];
        memset (lo, 0, sizeof (struct nbd_device));
        lo->magic     = NBD_DEV_MAGIC;
        for ( j = 0; j < NBD_MAXCONN; j++) {    /* PTB */
          struct nbd_slot * slot = & lo->slots[j];
          slot->lo = lo;
          slot->i  = j;
          NBD_INIT_LIST_HEAD(slot);
        }
        lo->blksize   = 1024;                   /* PTB 132 */
        lo->logblksize= 10;                     /* PTB */
        lo->bytesize  = 0x7fffffff;             /* PTB 132 */
        lo->size      = 0x7fffffff;             /* PTB (bytesizes/1024) */
        lo->nbd       = i;
        NBD_INIT_LIST_HEAD(lo);
        init_waitqueue_head(&lo->wq);
        for ( j = 0; j < NBD_MAXCONN; j++) {
          nbd_blksizes [i * NBD_MAXCONN + j] = lo->blksize;
          nbd_bytesizes[i * NBD_MAXCONN + j] = lo->bytesize;
          nbd_sizes    [i * NBD_MAXCONN + j] = lo->size;
        }

        //nbd_partitions[i << NBD_SHIFT].nr_sects = 2 * lo->size; 

    }

    NBD_DEBUG (3, "creating proc entry\n");
#if LINUX_VERSION_CODE >= LINUX_VERSION_2_3_30
    do {
      struct proc_dir_entry * res =
        create_proc_read_entry("nbdinfo", 0, NULL, &nbd_read_proc, NULL); 
      if (!res) {
        NBD_ALERT ("creation of proc entry failed\n");
        err = -EINVAL;
        NBD_DEBUG (3, "exited INVAL\n");
        return err;
      }
      // PTB additional write_proc entry in struct
      res->write_proc = &nbd_write_proc;
    } while(0);
#else
    do {
      struct proc_dir_entry * res = &nbd_proc_entry;
      proc_register (&proc_root, res);
      // PTB additional write_proc entry in struct
      res->write_proc = &nbd_write_proc;
    } while(0);
#endif

    NBD_DEBUG (3, "exited OK\n");
    return err;
}

#ifdef MODULE
void
cleanup_module (void)
{
    int i;

    NBD_DEBUG (3, "removing proc entry\n");
#if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
    proc_unregister (&proc_root, nbd_proc_entry.low_ino);
#else
    remove_proc_entry ("nbdinfo", &proc_root);
#endif

    for (i = 0; i < MAX_NBD ; i++) {
      struct nbd_device * lo = & nbd_dev[i];
      if (lo->blockmap) {
         kfree(lo->blockmap);
         lo->blockmap = NULL;
      }
      fsync_dev(MKDEV(MAJOR_NR,i<<NBD_SHIFT));
    }

    blk_cleanup_queue(BLK_DEFAULT_QUEUE(MAJOR_NR));
    blk_size[MAJOR_NR]           = NULL;
    read_ahead[MAJOR_NR]         = 0;

    if (unregister_blkdev (MAJOR_NR, "nbd") != 0) {
        NBD_ALERT ("cleanup_module failed\n");
    } else {
        NBD_INFO ("module cleaned up.\n");
    }
}
#endif

/* Compile line:

 *  gcc -O2 -D__KERNEL__ -DMODULE -xc -c nbd.c -o nbd.o
 *
 *  (possibly with -DMODVERSIONS also). PTB
 */
